from collections import OrderedDict
import concurrent.futures as cf
import numpy as np
import numba
from numba import types, typed
from numba.experimental import jitclass
import pywt
from pfb.wavelets import coeff_size, signal_size, dwt2d, idwt2d, copyT


@numba.njit
def create_dict(items):
    return {k: v for k,v in items}


def psi_band_maker(nx, ny, bases, nlevel):
    Nbasis = len(bases)
    sqrtNbasis = np.sqrt(Nbasis)
    dec_lo = {}
    dec_hi = {}
    rec_lo = {}
    rec_hi = {}
    sx = {}
    sy = {}
    spx = {}
    spy = {}
    Ntotx = {}
    Ntoty = {}
    ix = {}
    iy = {}
    Nxmax = 0
    Nymax = 0
    for wavelet in bases:
        if wavelet == 'self':
            continue
        wvlt = pywt.Wavelet(wavelet)
        dec_lo[wavelet] = np.array(wvlt.filter_bank[0])
        dec_hi[wavelet] = np.array(wvlt.filter_bank[1])
        rec_lo[wavelet] = np.array(wvlt.filter_bank[2])
        rec_hi[wavelet] = np.array(wvlt.filter_bank[3])

        max_level = pywt.dwt_max_level(np.minimum(nx, ny), wavelet)
        if nlevel > max_level:
            raise ValueError(f"The requested decomposition level {nlevel} "
                             "is not possible")

        # bookeeping
        N2Cx = {}
        N2Cy = {}
        Nx = nx
        Ny = ny
        Nxm = 0
        Nym = 0
        sx[wavelet] = typed.List()
        sy[wavelet] = typed.List()
        spx[wavelet] = typed.List()
        spy[wavelet] = typed.List()
        F = int(wavelet[-1])*2  # filter length
        for k in range(nlevel):
            Cx = coeff_size(Nx, F)
            Cy = coeff_size(Ny, F)
            N2Cx[k] = (signal_size(Cx, F), Cx)
            N2Cy[k] = (signal_size(Cy, F), Cy)
            Nxm += Cx
            Nym += Cy
            sx[wavelet].append(Cx)
            sy[wavelet].append(Cy)
            Nx = Cx + Cx%2
            Ny = Cy + Cy%2
            spx[wavelet].append(signal_size(Cx, F))
            spy[wavelet].append(signal_size(Cy, F))
        Nxm += Cx  # last approx coeffs
        Nym += Cy
        Ntotx[wavelet] = Nxm
        Ntoty[wavelet] = Nym
        Nxmax = np.maximum(Nxmax, Nxm)
        Nymax = np.maximum(Nymax, Nym)

        ix[wavelet] = {}
        iy[wavelet] = {}
        lowx = N2Cx[nlevel-1][1]
        lowy = N2Cy[nlevel-1][1]
        ix[wavelet][nlevel-1] = (lowx, 2*lowx)
        iy[wavelet][nlevel-1] = (lowy, 2*lowy)
        lowx *= 2
        lowy *= 2
        for k in reversed(range(nlevel-1)):
            highx = N2Cx[k][1]
            highy = N2Cy[k][1]
            ix[wavelet][k] = (lowx, lowx + highx)
            iy[wavelet][k] = (lowy, lowy + highy)
            lowx += highx
            lowy += highy

        ix[wavelet] = create_dict(tuple(ix[wavelet].items()))
        iy[wavelet] = create_dict(tuple(iy[wavelet].items()))

    # this avoids slow constructor
    ix = create_dict(tuple(ix.items()))
    iy = create_dict(tuple(iy.items()))
    dec_lo = create_dict(tuple(dec_lo.items()))
    dec_hi = create_dict(tuple(dec_hi.items()))
    rec_lo = create_dict(tuple(rec_lo.items()))
    rec_hi = create_dict(tuple(rec_hi.items()))
    sx = create_dict(tuple(sx.items()))
    sy = create_dict(tuple(sy.items()))
    spx = create_dict(tuple(spx.items()))
    spy = create_dict(tuple(spy.items()))
    Ntotx = create_dict(tuple(Ntotx.items()))
    Ntoty = create_dict(tuple(Ntoty.items()))

    # set up buffers
    alpha = np.zeros((Nymax, Nxmax))  # avoid destroying coeff in
    cbuff = np.zeros((Nxmax, Nymax))
    cbuffT = np.zeros((Nymax, Nxmax))
    image = np.zeros((nx, ny))

    return psi_band(alpha, image, cbuff, cbuffT,
                    ix, iy, sx, sy, spx, spy,
                    dec_lo, dec_hi, rec_lo, rec_hi,
                    Ntotx, Ntoty, Nxmax, Nymax, Nbasis,
                    sqrtNbasis, typed.List(bases), nlevel, nx, ny)


kv_ty = (types.unicode_type, types.ListType(numba.int64))
kv_ty2 = (numba.int64, types.UniTuple(types.int64, 2))
kv_ty3 = (types.unicode_type, numba.float64[:])
spec = OrderedDict()
spec['alpha'] = numba.float64[:,:]
spec['image'] = numba.float64[:,:]
spec['cbuff'] = numba.float64[:,:]
spec['cbuffT'] = numba.float64[:,:]
spec['ix'] = types.DictType(types.unicode_type, types.DictType(*kv_ty2))
spec['iy'] = types.DictType(types.unicode_type, types.DictType(*kv_ty2))
spec['sx'] = types.DictType(*kv_ty)
spec['sy'] = types.DictType(*kv_ty)
spec['spx'] = types.DictType(*kv_ty)
spec['spy'] = types.DictType(*kv_ty)
spec['dec_lo'] = types.DictType(*kv_ty3)
spec['dec_hi'] = types.DictType(*kv_ty3)
spec['rec_lo'] = types.DictType(*kv_ty3)
spec['rec_hi'] = types.DictType(*kv_ty3)
spec['Ntotx'] = types.DictType(types.unicode_type, numba.int64)
spec['Ntoty'] = types.DictType(types.unicode_type, numba.int64)
spec['Nxmax'] = numba.int64
spec['Nymax'] = numba.int64
spec['Nbasis'] = numba.int64
spec['sqrtNbasis'] = numba.float64
spec['bases'] = types.ListType(types.unicode_type)
spec['Nlevel'] = numba.int64
spec['Nx'] = numba.int64
spec['Ny'] = numba.int64
@jitclass(spec)
class psi_band(object):
    def __init__(self, alpha, image, cbuff, cbuffT,
                 ix, iy, sx, sy, spx, spy,
                 dec_lo, dec_hi, rec_lo, rec_hi,
                 Ntotx, Ntoty, Nxmax, Nymax, Nbasis,
                 sqrtNbasis, bases, Nlevel, Nx, Ny):

        self.alpha = alpha
        self.image = image
        self.cbuff = cbuff
        self.cbuffT = cbuffT
        self.ix = ix
        self.iy = iy
        self.sx = sx
        self.sy = sy
        self.spx = spx
        self.spy = spy
        self.dec_lo = dec_lo
        self.dec_hi = dec_hi
        self.rec_lo = rec_lo
        self.rec_hi = rec_hi
        self.Ntotx = Ntotx
        self.Ntoty = Ntoty
        self.Nxmax = Nxmax
        self.Nymax = Nymax
        self.Nbasis = Nbasis
        self.sqrtNbasis = sqrtNbasis
        self.bases = bases
        self.Nlevel = Nlevel
        self.Nx = Nx
        self.Ny = Ny

    def dot(self, x, alphao):
        '''
        signal to coeffs

        x       - (nx, ny) input signal
        alphao  - (nbasis, Nymax, Nxmax) per basis output coeffs
        '''

        alphao[...] = 0.0

        for i, wavelet in enumerate(self.bases):
            if wavelet=='self':
                copyT(x, alphao[i, 0:self.Ny, 0:self.Nx])
                # alphao[i, 0:self.Ny, 0:self.Nx] = x.T
                continue
            dec_lo = self.dec_lo[wavelet]
            dec_hi = self.dec_hi[wavelet]
            sx = self.sx[wavelet]
            sy = self.sy[wavelet]
            ix = self.ix[wavelet]
            iy = self.iy[wavelet]
            Ntotx = self.Ntotx[wavelet]
            Ntoty = self.Ntoty[wavelet]

            dwt2d(x,
                  alphao[i, 0:Ntoty, 0:Ntotx],
                  self.cbuff[0:Ntotx, 0:Ntoty],
                  self.cbuffT[0:Ntoty, 0:Ntotx],
                  ix, iy, sx, sy, dec_lo, dec_hi,
                  self.Nlevel)

        # alphao /= self.sqrtNbasis

        return alphao


    def hdot(self, alpha, xo):
        '''
        coeffs to signal

        alpha   - (nbasis, Nxmax, Nymax) per basis output coeffs
        xo      - (nx, ny) output signal
        '''
        xo[...] = 0.0  # accumulated
        for i, wavelet in enumerate(self.bases):
            if wavelet=='self':
                copyT(alpha[i, 0:self.Ny, 0:self.Nx], self.image)
                xo += self.image
                continue
            rec_lo = self.rec_lo[wavelet]
            rec_hi = self.rec_hi[wavelet]
            sx = self.sx[wavelet]
            sy = self.sy[wavelet]
            spx = self.spx[wavelet]
            spy = self.spy[wavelet]
            ix = self.ix[wavelet]
            iy = self.iy[wavelet]
            Ntotx = self.Ntotx[wavelet]
            Ntoty = self.Ntoty[wavelet]
            idwt2d(alpha[i, 0:Ntoty, 0:Ntotx],
                   self.image,
                   self.alpha[0:Ntoty, 0:Ntotx],
                   self.cbuff[0:Ntotx, 0:Ntoty],
                   self.cbuffT[0:Ntoty, 0:Ntotx],
                   ix, iy, sx, sy, spx, spy, rec_lo, rec_hi,
                   self.Nlevel)

            xo += self.image

        # xo /= self.sqrtNbasis

        return xo


def psi_dot_impl(x, alphao, psib, b, nthreads=1):
    numba.set_num_threads(nthreads)
    psib.dot(x, alphao)
    return b


def psi_hdot_impl(alpha, xo, psib, b, nthreads=1):
    numba.set_num_threads(nthreads)
    psib.hdot(alpha, xo)
    return b


class Psi(object):
    def __init__(self, nband, nx, ny, bases, nlevel, nthreads):

        self.psib = []
        for b in range(nband):
            self.psib.append(psi_band_maker(nx, ny, bases, nlevel))

        self.nband = nband
        self.nx = nx
        self.ny = ny
        self.nbasis = len(bases)
        self.nthreads = nthreads
        self.Nxmax = self.psib[0].Nxmax
        self.Nymax = self.psib[0].Nymax
        self.nthreads_per_band = np.maximum(1, nthreads//nband)

    def dot(self, x, alphao):
        '''
        image to coeffs
        '''
        futures = []
        with cf.ThreadPoolExecutor(max_workers=self.nband) as executor:
            for b in range(self.nband):
                f = executor.submit(psi_dot_impl, x[b], alphao[b], self.psib[b], b,
                                    nthreads=self.nthreads_per_band)
                futures.append(f)

            # wait for result
            for f in cf.as_completed(futures):
                b = f.result()

    def hdot(self, alpha, xo):
        '''
        coeffs to image
        '''
        futures = []
        with cf.ThreadPoolExecutor(max_workers=self.nband) as executor:
            for b in range(self.nband):
                f = executor.submit(psi_hdot_impl, alpha[b], xo[b], self.psib[b], b,
                                    nthreads=self.nthreads_per_band)
                futures.append(f)

            # wait for result
            for f in cf.as_completed(futures):
                b = f.result()



import numpy as np
from functools import partial
from pfb.opt.fista import fista
from pfb.operators.psf import PSF
from pfb.utils.fits import save_fits
from pfb.opt.power_method import power_method
from pfb.utils import logging as pfb_logging
log = pfb_logging.get_logger('NNLS')


def resid_func(x, dirty, hessian, mask, beam, wsum):
    """
    Returns the unattenuated residual
    """
    residual = dirty - hessian(mask(beam(x)))/wsum
    residual_mfs = np.sum(residual, axis=0)
    residual = residual
    return residual, residual_mfs


def value_and_grad(x, dirty, psfo, mask, beam):
    convmod = mask(beam(psfo.convolve(mask(beam(x)))))
    attdirty = mask(beam(dirty))
    return np.vdot(x, convmod - 2*attdirty), 2*convmod - 2*attdirty

def prox(x):
    x[x<0] = 0.0
    return x

def nnls(psf, model, residual, mask=None, beam_image=None,
         hessian=None, wsum=None, gamma=0.95,
         hdr=None, hdr_mfs=None, outfile=None,
         nthreads=1, maxit=1, tol=1e-3,
         pmtol=1e-5, pmmaxit=50, pmverbose=1,
         ftol=1e-5, fmaxit=250, fverbose=3):

    if len(residual.shape) > 3:
        raise ValueError("Residual must have shape (nband, nx, ny)")

    nband, nx, ny = residual.shape

    if beam_image is None:
        def beam(x): return x
    else:
        try:
            assert beam.shape == (nband, nx, ny)
            def beam(x): return beam_image * x
        except BaseException:
            raise ValueError("Beam has incorrect shape")

    if mask is None:
        def mask(x): return x
    else:
        try:
            if mask.ndim == 2:
                assert mask.shape == (nx, ny)
                def mask(x): return mask[None] * x
            elif mask.ndim == 3:
                assert mask.shape == (1, nx, ny)
                def mask(x): return mask * x
            else:
                raise ValueError
        except BaseException:
            raise ValueError("Mask has incorrect shape")

    # PSF operator
    psfo = PSF(psf, residual.shape, nthreads=nthreads)

    residual_mfs = np.sum(residual, axis=0)
    residual = mask(beam(residual))
    rmax = np.abs(residual_mfs).max()
    rms = np.std(residual_mfs)

    if hessian is None:
        hessian = psfo.convolve
        wsum = 1

    def hess(x):
        return mask(beam(psfo.convolve(mask(beam(x)))))

    beta, betavec = power_method(hess, residual.shape, tol=pmtol,
                                 maxit=pmmaxit, verbosity=pmverbose)

    if model.any():
        dirty = residual + hessian(mask(beam(model)))/wsum
    else:
        dirty = residual

    for i in range(maxit):
        fprime = partial(value_and_grad, dirty=residual,
                         psfo=psfo, mask=mask, beam=beam)

        x = fista(np.zeros_like(model), beta,
                  fprime, prox, tol=ftol, maxit=fmaxit,
                  verbosity=fverbose)

        modelp = model.copy()
        model += gamma * x

        residual, residual_mfs = resid_func(model, dirty, hessian, mask, beam, wsum)
        model_mfs = np.mean(model, axis=0)

        # check stopping criteria
        rmax = np.abs(residual_mfs).max()
        rms = np.std(residual_mfs)
        eps = np.linalg.norm(model - modelp)/np.linalg.norm(model)

        log.info("Iter %i: peak residual = %f, rms = %f, eps = %f" % (
              i+1, rmax, rms, eps))

        # save current iteration
        if outfile is not None:
            assert hdr is not None
            assert hdr_mfs is not None

            save_fits(outfile + str(i + 1) + '_NNLS_model_mfs.fits',
                        model_mfs, hdr_mfs)

            save_fits(outfile + str(i + 1) + '_NNLS_model.fits',
                        model, hdr)

            save_fits(outfile + str(i + 1) + '_NNLS_residual_mfs.fits',
                        residual_mfs, hdr_mfs)

        if eps < tol:
            log.info("Success, convergence after %i iterations" % (i+1))
            break

    return model


# flake8: noqa
import concurrent.futures as cf
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('SARA')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.sara)
def sara(**kw):
    '''
    Deconvolution using SARA regularisation
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    OmegaConf.set_struct(opts, True)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/sara_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    from pfb.utils.naming import xds_from_url, get_opts

    basename = opts.output_filename
    fits_oname = f'{opts.fits_output_folder}/{oname}'
    dds_name = f'{basename}_{opts.suffix}.dds'

    ti = time.time()
    _sara(**opts)

    dds, dds_list = xds_from_url(dds_name)

    if opts.fits_mfs or opts.fits:
        from daskms.fsspec_store import DaskMSStore
        from pfb.utils.fits import dds2fits
        # get the psfpars for the mfs cube
        dds_store = DaskMSStore(dds_name)
        if '://' in dds_store.url:
            protocol = dds_store.url.split('://')[0]
        else:
            protocol = 'file'
        psfpars_mfs = get_opts(dds_store.url,
                               protocol,
                               name='psfparsn_mfs.pkl')
        log.info(f"Writing fits files to {fits_oname}_{opts.suffix}")

        dds2fits(dds_list,
                'RESIDUAL',
                f'{fits_oname}_{opts.suffix}',
                norm_wsum=True,
                nthreads=opts.nthreads,
                do_mfs=opts.fits_mfs,
                do_cube=opts.fits_cubes,
                psfpars_mfs=psfpars_mfs)
        log.info('Done writing RESIDUAL')
        dds2fits(dds_list,
                'MODEL',
                f'{fits_oname}_{opts.suffix}',
                norm_wsum=False,
                nthreads=opts.nthreads,
                do_mfs=opts.fits_mfs,
                do_cube=opts.fits_cubes,
                psfpars_mfs=psfpars_mfs)
        log.info('Done writing MODEL')
        dds2fits(dds_list,
                'UPDATE',
                f'{fits_oname}_{opts.suffix}',
                norm_wsum=False,
                nthreads=opts.nthreads,
                do_mfs=opts.fits_mfs,
                do_cube=opts.fits_cubes,
                psfpars_mfs=psfpars_mfs)
        log.info('Done writing UPDATE')
        # try:
        #     dds2fits(dds_list,
        #          'MPSF',
        #          f'{fits_oname}_{opts.suffix}',
        #          norm_wsum=False,
        #          nthreads=opts.nthreads,
        #          do_mfs=opts.fits_mfs,
        #          do_cube=opts.fits_cubes)
        #     log.info('Done writing MPSF')
        # except Exception as e:
        #     print(e)

    from numba import threading_layer
    log.info(f"Numba use the {threading_layer()} threading layer")
    log.info(f"All done after {time.time() - ti}s")


def _sara(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    from functools import partial
    import numpy as np
    import xarray as xr
    import numexpr as ne
    from pfb.utils.fits import set_wcs, save_fits
    from pfb.utils.naming import xds_from_url
    from pfb.opt.power_method import power_method
    from pfb.opt.primal_dual import primal_dual_optimised as primal_dual
    from pfb.utils.misc import l1reweight_func
    from pfb.operators.hessian import hess_psf
    from pfb.operators.psi import Psi
    from pfb.operators.gridder import compute_residual
    from copy import deepcopy
    from ducc0.misc import thread_pool_size
    from pfb.prox.prox_21m import prox_21m_numba as prox_21
    from pfb.utils.modelspec import fit_image_cube, eval_coeffs_to_slice

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    dds_name = f'{basename}_{opts.suffix}.dds'
    dds, dds_list = xds_from_url(dds_name)

    if dds[0].corr.size > 1:
        log.error_and_raise("Joint polarisation deconvolution not "
                            "yet supported for sara algorithm",
                            NotImplementedError)

    nx, ny = dds[0].x.size, dds[0].y.size

    nx_psf, ny_psf = dds[0].x_psf.size, dds[0].y_psf.size
    lastsize = ny_psf
    freq_out = []
    time_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
        time_out.append(ds.time_out)
    freq_out = np.unique(np.array(freq_out))
    time_out = np.unique(np.array(time_out))
    if time_out.size > 1:
        log.error_and_raise('Only static models currently supported',
                            NotImplementedError)

    nband = freq_out.size

    # drop_vars after access to avoid duplicates in memory
    # and avoid unintentional side effects?
    output_type = dds[0].DIRTY.dtype
    if 'RESIDUAL' in dds[0]:
        residual = np.stack([ds.RESIDUAL.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('DIRTY') for ds in dds]
        dds = [ds.drop_vars('RESIDUAL') for ds in dds]
    else:
        residual = np.stack([ds.DIRTY.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('DIRTY') for ds in dds]
    if 'MODEL' in dds[0]:
        model = np.stack([ds.MODEL.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('MODEL') for ds in dds]
    else:
        model = np.zeros((nband, nx, ny))
    if 'UPDATE' in dds[0]:
        update = np.stack([ds.UPDATE.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('UPDATE') for ds in dds]
    else:
        update = np.zeros((nband, nx, ny))
    abspsf = np.stack([np.abs(ds.PSFHAT.values[0]) for ds in dds], axis=0)
    dds = [ds.drop_vars('PSFHAT') for ds in dds]
    beam = np.stack([ds.BEAM.values[0] for ds in dds], axis=0)
    wsums = np.stack([ds.WSUM.values[0] for ds in dds], axis=0)
    fsel = wsums > 0  # keep track of empty bands

    wsum = np.sum(wsums)
    wsums /= wsum
    abspsf /= wsum
    residual /= wsum
    residual_mfs = np.sum(residual, axis=0)
    model_mfs = np.mean(model[fsel], axis=0)

    # for intermediary results
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq, casambm=False)
    if 'niters' in dds[0].attrs:
        iter0 = dds[0].niters
    else:
        iter0 = 0

    # Allow calling with pd_tol as float
    try:
        ntol = len(opts.pd_tol)
        pd_tol = opts.pd_tol
    except TypeError:
        assert isinstance(opts.pd_tol, float)
        ntol = 1
        pd_tol = [opts.pd_tol]
    niters = opts.niter
    if ntol <= niters:
        pd_tolf = pd_tol[-1]
        pd_tol += [pd_tolf]*niters  # no harm in too many

    # image space hessian
    # pre-allocate arrays for doing FFT's
    real_type = 'f8'
    complex_type = 'c16'
    precond = hess_psf(nx, ny, abspsf,
                       beam=beam,
                       eta=opts.eta*wsums,
                       nthreads=opts.nthreads,
                       cgtol=opts.cg_tol,
                       cgmaxit=opts.cg_maxit,
                       cgverbose=opts.cg_verbose,
                       cgrf=opts.cg_report_freq,
                       taper_width=np.minimum(int(0.1*nx), 32))

    if opts.hess_norm is None:
        # if the grid worker had been rerun hess_norm won't be in attrs
        if 'hess_norm' in dds[0].attrs:
            hess_norm = dds[0].hess_norm
            log.info(f"Using previously estimated hess_norm of {hess_norm:.3e}")
        else:
            log.info("Finding spectral norm of Hessian approximation")
            hess_norm, hessbeta = power_method(
                                            precond.dot, (nband, nx, ny),
                                            tol=opts.pm_tol,
                                            maxit=opts.pm_maxit,
                                            verbosity=opts.pm_verbose,
                                            report_freq=opts.pm_report_freq)
            # inflate slightly for stability
            hess_norm *= 1.05
    else:
        hess_norm = opts.hess_norm
        log.info(f"Using provided hess-norm of = {hess_norm:.3e}")

    log.info("Setting up dictionary")
    bases = tuple(opts.bases.split(','))
    nbasis = len(bases)
    psi = Psi(nband, nx, ny, bases, opts.nlevels, opts.nthreads)
    Nxmax = psi.Nxmax
    Nymax = psi.Nymax

    log.info(f"Using {psi.nthreads_per_band} numba threads for each band")
    log.info(f"Using {thread_pool_size()} threads for gridding")

    # number of frequency basis functions
    if opts.nbasisf is None:
        nbasisf = int(np.sum(fsel))
    else:
        nbasisf = opts.nbasisf
    log.info(f"Using {nbasisf} frequency basis functions")

    # a value less than zero turns L1 reweighting off
    # we'll start on convergence or at the iteration
    # indicated by l1-reweight-from, whichever comes first
    l1_reweight_from = opts.l1_reweight_from
    l1reweight_active = False
    # we need an array to put the components in for reweighting
    outvar = np.zeros((nband, nbasis, Nymax, Nxmax), dtype=real_type)

    dual = np.zeros((nband, nbasis, Nymax, Nxmax), dtype=residual.dtype)
    if l1_reweight_from == 0:
        log.info('Initialising with L1 reweighted')
        if not update.any():
            log.error_and_raise("Cannot reweight before any updates have been performed",
                                ValueError)
        psi.dot(update, outvar)
        tmp = np.sum(outvar, axis=0)
        # exclude zeros from padding DWT's
        # rms_comps = np.std(tmp[tmp!=0])
        # log.info(f'rms_comps updated to {rms_comps}')
        # per basis rms_comps
        rms_comps = np.ones((nbasis,), dtype=float)
        for i, base in enumerate(bases):
            tmpb = tmp[i]
            rms_comps[i] = np.std(tmpb[tmpb!=0])
            log.info(f'rms_comps for base {base} is {rms_comps[i]}')
        reweighter = partial(l1reweight_func,
                             psiH=psi.dot,
                             outvar=outvar,
                             rmsfactor=opts.rmsfactor,
                             rms_comps=rms_comps,
                             alpha=opts.alpha)
        l1weight = reweighter(model)
        l1reweight_active = True
    else:
        l1weight = np.ones((nbasis, Nymax, Nxmax), dtype=residual.dtype)
        reweighter = None
        l1reweight_active = False

    if opts.rms_outside_model and model.any():
        rms_mask = model_mfs == 0
        rms = np.std(residual_mfs[rms_mask])
    else:
        rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()
    best_rms = rms
    best_rmax = rmax
    best_model = model.copy()
    diverge_count = 0
    eps = 1.0
    write_futures = None
    log.info(f"Iter {iter0}: peak residual = {rmax:.3e}, rms = {rms:.3e}")
    if opts.skip_model:
        mrange = []
    else:
        mrange = range(iter0, iter0 + opts.niter)
    for k in mrange:
        log.info('Solving for update')
        residual *= beam  # avoid copy
        update = precond.idot(residual,
                              mode=opts.hess_approx,
                              x0=update if update.any() else None)
        update_mfs = np.mean(update, axis=0)
        save_fits(update_mfs,
                  fits_oname + f'_{opts.suffix}_update_{k+1}.fits',
                  hdr_mfs)

        modelp = deepcopy(model)
        xtilde = model + opts.gamma * update
        grad21 = lambda x: -precond.dot(xtilde - x)/opts.gamma
        if iter0 == 0:
            lam = opts.init_factor * opts.rmsfactor * rms
        else:
            lam = opts.rmsfactor*rms
        log.info(f'Solving for model with lambda = {lam}')
        model, dual = primal_dual(model,
                                  dual,
                                  lam,
                                  psi.hdot,
                                  psi.dot,
                                  hess_norm,
                                  prox_21,
                                  l1weight,
                                  reweighter,
                                  grad21,
                                  nu=nbasis,
                                  positivity=opts.positivity,
                                  tol=pd_tol[k-iter0],
                                  maxit=opts.pd_maxit,
                                  verbosity=opts.pd_verbose,
                                  report_freq=opts.pd_report_freq,
                                  gamma=opts.gamma)

        # write component model
        log.info(f"Writing model to {basename}_{opts.suffix}_model.mds")
        try:
            coeffs, Ix, Iy, expr, params, texpr, fexpr = \
                fit_image_cube(time_out,
                               freq_out[fsel],
                               model[None, fsel, :, :],
                               wgt=wsums[None, fsel],
                               nbasisf=nbasisf,
                               method='Legendre',
                               sigmasq=1e-6)
            # save interpolated dataset
            data_vars = {
                'coefficients': (('par', 'comps'), coeffs),
            }
            coords = {
                'location_x': (('x',), Ix),
                'location_y': (('y',), Iy),
                # 'shape_x':,
                'params': (('par',), params),  # already converted to list
                'times': (('t',), time_out),  # to allow rendering to original grid
                'freqs': (('f',), freq_out)
            }
            mattrs = {
                'spec': 'genesis',
                'cell_rad_x': cell_rad,
                'cell_rad_y': cell_rad,
                'npix_x': nx,
                'npix_y': ny,
                'texpr': texpr,
                'fexpr': fexpr,
                'center_x': dds[0].x0,
                'center_y': dds[0].y0,
                'flip_u': dds[0].flip_u,
                'flip_v': dds[0].flip_v,
                'flip_w': dds[0].flip_w,
                'ra': dds[0].ra,
                'dec': dds[0].dec,
                'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
                'parametrisation': expr  # already converted to str
            }
            for key, val in opts.items():
                if key == 'pd_tol':
                    mattrs[key] = pd_tolf
                else:
                    mattrs[key] = val

            coeff_dataset = xr.Dataset(data_vars=data_vars,
                               coords=coords,
                               attrs=mattrs)
            coeff_dataset.to_zarr(f"{basename}_{opts.suffix}_model.mds",
                                  mode='w')

            for b in range(nband):
                model[b] = eval_coeffs_to_slice(
                        time_out[0],
                        freq_out[b],
                        coeffs,
                        Ix, Iy,
                        expr,
                        params,
                        texpr,
                        fexpr,
                        nx, ny,
                        cell_rad, cell_rad,
                        dds[0].x0, dds[0].y0,
                        nx, ny,
                        cell_rad, cell_rad,
                        dds[0].x0, dds[0].y0
                )
        except Exception as e:
            log.info(f"Exception {e} raised during model fit .")

        model_mfs = np.mean(model[fsel], axis=0)
        save_fits(model_mfs,
                  fits_oname + f'_{opts.suffix}_model_{k+1}.fits',
                  hdr_mfs)

        # make sure write futures have finished
        if write_futures is not None:
            cf.wait(write_futures)

        log.info(f'Computing residual')
        write_futures = []
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            resid, fut = compute_residual(
                                    ds_name,
                                    nx, ny,
                                    cell_rad, cell_rad,
                                    ds_name,
                                    model[b][None, :, :],  # add corr axis
                                    nthreads=opts.nthreads,
                                    epsilon=opts.epsilon,
                                    do_wgridding=opts.do_wgridding,
                                    double_accum=opts.double_accum,
                                    verbosity=opts.verbosity)
            write_futures.append(fut)
            residual[b] = resid[0]  # remove corr axis

        residual /= wsum
        residual_mfs = np.sum(residual, axis=0)
        save_fits(residual_mfs,
                  fits_oname + f'_{opts.suffix}_residual_{k+1}.fits',
                  hdr_mfs)
        rmsp = rms
        if opts.rms_outside_model:
            rms_mask = model_mfs == 0
            rms = np.std(residual_mfs[rms_mask])
        else:
            rms = np.std(residual_mfs)
        rmaxp = rmax
        rmax = np.abs(residual_mfs).max()
        eps = np.linalg.norm(model - modelp)/np.linalg.norm(model)

        # what to base this on?
        if rms < best_rms:
            best_rms = rms
            best_rmax = rmax
            best_model = model.copy()

        # these are not updated in compute_residual
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            ds['UPDATE'] = (('corr', 'x', 'y'), update[b][None, :, :])
            # don't write unecessarily
            for var in ds.data_vars:
                if var != 'UPDATE':
                    ds = ds.drop_vars(var)

            if (model==best_model).all():
                ds['MODEL_BEST'] = (('corr', 'x', 'y'), best_model[b][None, :, :])

            attrs = {}
            attrs['rms'] = best_rms
            attrs['rmax'] = best_rmax
            attrs['niters'] = k+1
            attrs['hess_norm'] = hess_norm
            ds = ds.assign_attrs(**attrs)


            with cf.ThreadPoolExecutor(max_workers=1) as executor:
                fut = executor.submit(ds.to_zarr, ds_name, mode='a')
                write_futures.append(fut)

            # ds.to_zarr(ds_name, mode='a')


        log.info(f"Iter {k+1}: peak residual = {rmax:.3e}, "
              f"rms = {rms:.3e}, eps = {eps:.3e}")

        if eps < opts.tol:
            # do not converge prematurely
            if l1_reweight_from > 0 and not l1reweight_active:  # only happens once
                # start reweighting
                l1_reweight_from = k+1 - iter0
                l1reweight_active = True
            else:
                log.info(f"Converged after {k+1} iterations.")
                break

        if (k+1 - iter0 >= l1_reweight_from) and (k+1 - iter0 < opts.niter):
            log.info('Computing L1 weights')
            psi.dot(update, outvar)
            tmp = np.sum(outvar, axis=0)
            # exclude zeros from padding DWT's
            # rms_comps = np.std(tmp[tmp!=0])
            rms_comps = np.ones((nbasis,), dtype=float)
            # log.info(f'rms_comps updated to {rms_comps}')
            for i, base in enumerate(bases):
                tmpb = tmp[i]
                rms_comps[i] = np.std(tmpb[tmpb!=0])
                log.info(f'rms_comps for base {base} is {rms_comps[i]}')
            reweighter = partial(l1reweight_func,
                                 psiH=psi.dot,
                                 outvar=outvar,
                                 rmsfactor=opts.rmsfactor,
                                 rms_comps=rms_comps,
                                 alpha=opts.alpha)
            l1weight = reweighter(model)
            l1reweight_active = True

        if (rms > rmsp) and (rmax > rmaxp):
            diverge_count += 1
            if diverge_count > opts.diverge_count:
                log.info("Algorithm is diverging. Terminating.")
                break

    # make sure write futures have finished
    cf.wait(write_futures)

    return



import numpy as np
import sympy as sm
from sympy.physics.quantum import TensorProduct
from sympy.utilities.lambdify import lambdify
from numba import njit
from numba.core import types
import string

def jones_to_mueller(gp, gq):
    shape = gp.shape
    rem_shape = shape[2:]
    i0 = string.ascii_lowercase.index('m')
    rem_idx = string.ascii_lowercase[i0:i0+len(rem_shape)]
    idxp = 'ij' + rem_idx
    idxq = 'kl' + rem_idx
    idxo = 'ikjl' + rem_idx
    out_shape = (4,4) + rem_shape
    return np.einsum(f'{idxp},{idxq}->{idxo}', gp, np.conjugate(gq)).reshape(*out_shape)

def mueller_to_stokes(mueller, poltype='linear'):
    '''
    Convert a Mueller matrix into a diagonal Stokes matrix.
    '''
    if poltype=='linear':
        T = np.array([[1.0, 1.0, 0, 0],
                      [0, 0, 1.0, 1.0j],
                      [0, 0, 1.0, -1.0j],
                      [1.0, -1.0, 0, 0]])
    elif poltype=='circular':
        T = np.array([[1.0, 0, 0, 1.0],
                      [0, 1.0, 1.0j, 0],
                      [0, 1.0, -1.0j, 0],
                      [1.0, 0, 0, -1.0]])
    else:
        raise ValueError(f'Unknown poltype {poltype}')
    shape = mueller.shape
    rem_shape = shape[2:]
    i0 = string.ascii_lowercase.index('m')
    rem_idx = string.ascii_lowercase[i0:i0+len(rem_shape)]
    idxl = 'ij' + rem_idx
    idxr = 'ji'
    idxo = 'i' + rem_idx
    return np.einsum(f'{idxl},{idxr}->{idxo}', mueller, T).real

def corr_to_stokes(x, wsum=1.0, axis=0, poltype='linear'):
    '''
    x = [I+Q, U+1jV, U-1jV, I-Q]
    out = [I, Q, U, V]
    '''
    if poltype.lower() != 'linear':
        raise NotImplementedError("Only linear polarisation is implemented")
    if x.shape[axis] != 4:
        raise ValueError(f"Expected 4 polarisation products, got {x.shape[axis]}")
    dirtyI = (np.take(x, 0, axis=axis) + np.take(x, 3, axis=axis)).real/wsum
    dirtyQ = (np.take(x, 0, axis=axis) - np.take(x, 3, axis=axis)).real/wsum
    dirtyU = (np.take(x, 1, axis=axis) + np.take(x, 2, axis=axis)).real/wsum
    dirtyV = (np.take(x, 1, axis=axis) - np.take(x, 2, axis=axis)).imag/wsum
    return np.stack((dirtyI, dirtyQ, dirtyU, dirtyV), axis=axis)

def stokes_to_corr(x, axis=0, poltype='linear'):
    '''
    x = [I, Q, U, V]
    out = [I+Q, U+1jV, U-1jV, I-Q]
    '''
    if poltype.lower() != 'linear':
        raise NotImplementedError("Only linear polarisation is implemented")
    if x.shape[axis] != 4:
        raise ValueError(f"Expected 4 polarisation products, got {x.shape[axis]}")
    dirty0 = np.take(x, 0, axis=axis) + np.take(x, 1, axis=axis)
    dirty1 = np.take(x, 2, axis=axis) + 1j*np.take(x, 3, axis=axis)
    dirty2 = np.take(x, 2, axis=axis) - 1j*np.take(x, 3, axis=axis)
    dirty3 = np.take(x, 0, axis=axis) - np.take(x, 1, axis=axis)
    return np.stack((dirty0, dirty1, dirty2, dirty3), axis=axis)


def stokes_funcs(data, jones, product, pol, nc):
    if isinstance(product, types.StringLiteral):
        product = product.literal_value
    if isinstance(pol, types.StringLiteral):
        pol = pol.literal_value
    if isinstance(nc, types.StringLiteral):
        nc = nc.literal_value

    # set up symbolic expressions
    gp00, gp10, gp01, gp11 = sm.symbols("gp00 gp10 gp01 gp11", real=False)
    gq00, gq10, gq01, gq11 = sm.symbols("gq00 gq10 gq01 gq11", real=False)
    w0, w1, w2, w3 = sm.symbols("W0 W1 W2 W3", real=True)
    v00, v10, v01, v11 = sm.symbols("v00 v10 v01 v11", real=False)

    # Jones matrices
    Gp = sm.Matrix([[gp00, gp01],[gp10, gp11]])
    Gq = sm.Matrix([[gq00, gq01],[gq10, gq11]])

    # Mueller matrix (row major form)
    Mpq = TensorProduct(Gp, Gq.conjugate())
    Mpqinv = TensorProduct(Gp.inv(), Gq.conjugate().inv())

    # inverse noise covariance
    Sinv = sm.Matrix([[w0, 0, 0, 0],
                      [0, w1, 0, 0],
                      [0, 0, w2, 0],
                      [0, 0, 0, w3]])
    S = Sinv.inv()

    # visibilities
    Vpq = sm.Matrix([[v00], [v01], [v10], [v11]])

    # Full Stokes to corr operator
    # Is this the only difference between linear and circular pol?
    # What about paralactic angle rotation?
    if pol == 'linear':
        T = sm.Matrix([[1.0, 1.0, 0, 0],
                       [0, 0, 1.0, 1.0j],
                       [0, 0, 1.0, -1.0j],
                       [1.0, -1.0, 0, 0]])
    elif pol == 'circular':
        T = sm.Matrix([[1.0, 0, 0, 1.0],
                       [0, 1.0, 1.0j, 0],
                       [0, 1.0, -1.0j, 0],
                       [1.0, 0, 0, -1.0]])
    Tinv = T.inv()

    # Full Stokes weights
    W = T.H * Mpq.H * Sinv * Mpq * T
    Winv = Tinv * Mpqinv * S * Mpqinv.H * Tinv.H

    # Full Stokes coherencies
    C = Winv * (T.H * (Mpq.H * (Sinv * Vpq)))
    # Only keep diagonal of weights
    W = W.diagonal().T  # diagonal() returns row vector

    # this should ensure that outputs are always ordered as
    # [I, Q, U, V]
    i = ()
    if 'I' in product:
        i += (0,)

    if 'Q' in product:
        i += (1,)
        if pol == 'circular' and nc == '2':
            raise ValueError("Q is not available in circular polarisation with 2 correlations")

    if 'U' in product:
        i += (2,)
        if pol == 'linear' and nc == '2':
            raise ValueError("U is not available in linear polarisation with 2 correlations")
        elif pol == 'circular' and nc == '2':
            raise ValueError("U is not available in circular polarisation with 2 correlations")

    if 'V' in product:
        i += (3,)
        if pol == 'linear' and nc == '2':
            raise ValueError("V is not available in linear polarisation with 2 correlations")

    remprod = product.strip('IQUV')
    if len(remprod):
        raise ValueError(f"Unknown polarisation product {remprod}")

    if jones.ndim == 6:  # Full mode
        Wsymb = lambdify((gp00, gp01, gp10, gp11,
                          gq00, gq01, gq10, gq11,
                          w0, w1, w2, w3),
                          sm.simplify(W[i,0]))
        Wjfn = njit(nogil=True, inline='always')(Wsymb)


        Dsymb = lambdify((gp00, gp01, gp10, gp11,
                          gq00, gq01, gq10, gq11,
                          w0, w1, w2, w3,
                          v00, v01, v10, v11),
                          sm.simplify(C[i,0]))
        Djfn = njit(nogil=True, inline='always')(Dsymb)

        @njit(nogil=True, inline='always')
        def wfunc(gp, gq, W):
            gp00 = gp[0,0]
            gp01 = gp[0,1]
            gp10 = gp[1,0]
            gp11 = gp[1,1]
            gq00 = gq[0,0]
            gq01 = gq[0,1]
            gq10 = gq[1,0]
            gq11 = gq[1,1]
            W00 = W[0]
            W01 = W[1]
            W10 = W[2]
            W11 = W[3]
            return Wjfn(gp00, gp01, gp10, gp11,
                        gq00, gq01, gq10, gq11,
                        W00, W01, W10, W11).real.ravel()

        @njit(nogil=True, inline='always')
        def vfunc(gp, gq, W, V):
            gp00 = gp[0,0]
            gp01 = gp[0,1]
            gp10 = gp[1,0]
            gp11 = gp[1,1]
            gq00 = gq[0,0]
            gq01 = gq[0,1]
            gq10 = gq[1,0]
            gq11 = gq[1,1]
            W00 = W[0]
            W01 = W[1]
            W10 = W[2]
            W11 = W[3]
            V00 = V[0]
            V01 = V[1]
            V10 = V[2]
            V11 = V[3]
            return Djfn(gp00, gp01, gp10, gp11,
                        gq00, gq01, gq10, gq11,
                        W00, W01, W10, W11,
                        V00, V01, V10, V11).ravel()

    elif jones.ndim == 5:  # DIAG mode
        W = W.subs(gp10, 0)
        W = W.subs(gp01, 0)
        W = W.subs(gq10, 0)
        W = W.subs(gq01, 0)
        C = C.subs(gp10, 0)
        C = C.subs(gp01, 0)
        C = C.subs(gq10, 0)
        C = C.subs(gq01, 0)

        Wsymb = lambdify((gp00, gp11,
                          gq00, gq11,
                          w0, w1, w2, w3),
                          sm.simplify(W[i,0]))
        Wjfn = njit(nogil=True, inline='always')(Wsymb)


        Dsymb = lambdify((gp00, gp11,
                          gq00, gq11,
                          w0, w1, w2, w3,
                          v00, v01, v10, v11),
                          sm.simplify(C[i,0]))
        Djfn = njit(nogil=True, inline='always')(Dsymb)

        if nc == '4':
            @njit(nogil=True, inline='always')
            def wfunc(gp, gq, W):
                gp00 = gp[0]
                gp11 = gp[1]
                gq00 = gq[0]
                gq11 = gq[1]
                W00 = W[0]
                W01 = W[1]
                W10 = W[2]
                W11 = W[3]
                return Wjfn(gp00, gp11,
                            gq00, gq11,
                            W00, W01, W10, W11).real.ravel()

            @njit(nogil=True, inline='always')
            def vfunc(gp, gq, W, V):
                gp00 = gp[0]
                gp11 = gp[1]
                gq00 = gq[0]
                gq11 = gq[1]
                W00 = W[0]
                W01 = W[1]
                W10 = W[2]
                W11 = W[3]
                V00 = V[0]
                V01 = V[1]
                V10 = V[2]
                V11 = V[3]
                return Djfn(gp00, gp11,
                            gq00, gq11,
                            W00, W01, W10, W11,
                            V00, V01, V10, V11).ravel()
        elif nc == '2':
            @njit(nogil=True, inline='always')
            def wfunc(gp, gq, W):
                gp00 = gp[0]
                gp11 = gp[1]
                gq00 = gq[0]
                gq11 = gq[1]
                W00 = W[0]
                W01 = 1.0
                W10 = 1.0
                W11 = W[-1]
                return Wjfn(gp00, gp11,
                            gq00, gq11,
                            W00, W01, W10, W11).real.ravel()

            @njit(nogil=True, inline='always')
            def vfunc(gp, gq, W, V):
                gp00 = gp[0]
                gp11 = gp[1]
                gq00 = gq[0]
                gq11 = gq[1]
                W00 = W[0]
                W01 = 1.0
                W10 = 1.0
                W11 = W[-1]
                V00 = V[0]
                V01 = 0j
                V10 = 0j
                V11 = V[-1]
                return Djfn(gp00, gp11,
                            gq00, gq11,
                            W00, W01, W10, W11,
                            V00, V01, V10, V11).ravel()
        else:
            raise ValueError(f"Selected product is only available from 2 or 4"
                             f"correlation data while you have ncorr={nc}.")


    else:
        raise ValueError(f"Jones term has incorrect number of dimensions")

    return vfunc, wfunc



"""
Logging utilities for PFB-Imaging using standard Python logging with Rich formatting.

This module provides a drop-in replacement for pyscilog with enhanced
formatting using the Rich library for better console output.
"""

import logging
import time
from pathlib import Path
from typing import Optional, Dict, Any, Type, Union
from functools import wraps

from rich.console import Console
from rich.logging import RichHandler
from rich.traceback import install as install_rich_traceback
from rich.table import Table, Column
from rich.panel import Panel
from rich.text import Text

rich_console = Console()

install_rich_traceback(console=rich_console, show_locals=False)

class PFBLogger(logging.Logger):
    """
    Enhanced logger class that provides pyscilog-compatible interface
    with Rich formatting capabilities.
    """

    def error_and_raise(
        self,
        message: str,
        exception_type: Type[Exception] = Exception,
        *args,
        **kwargs
    ) -> None:
        """
        Log an error message and raise an exception.

        This provides compatibility with pyscilog's error_and_raise method.

        Args:
            message: The error message to log and include in the exception
            exception_type: The type of exception to raise (default: Exception)
            *args: Additional arguments for the logger
            **kwargs: Additional keyword arguments for the logger
        """

        bold_message = f"[bold red]{message}[/bold red]"
        self.error(bold_message, *args, **kwargs)

        raise exception_type(message)

# Any logger created hereafter will be an instance of PFBLogger.
logging.setLoggerClass(PFBLogger)

class LoggingManager:
    """
    Global logging manager that handles initialization and configuration.
    """

    def __init__(self, app_name: str, log_level: Union[str, int] = logging.INFO) -> None:
        """
        Initialize the logging system.

        Args:
            app_name: Name of the application
            log_level: Logging level of the default console handler.
        """
        self._app_name = app_name
        self._loggers: Dict[str, PFBLogger] = {}
        self._log_files: Dict[str, str] = {}
        self._console = rich_console
        self._root_logger = logging.getLogger(self._app_name)

        # Set up the root logger - this should always be set to DEBUG level to ensure that all
        # log messages are dispatched to the relevant handlers.
        self._root_logger.setLevel(logging.DEBUG)

        # Remove existing handlers to avoid duplicates
        for handler in self._root_logger.handlers:
            self._root_logger.removeHandler(handler)

        # Create console handler with Rich formatting. The RichHandler has limited configuration
        # options. If needed, it is possible to subclass it and modify its methods.
        console_handler = RichHandler(
            console=self._console,
            show_time=False,
            show_level=False,
            show_path=False,
            rich_tracebacks=True,
            tracebacks_show_locals=True,
            markup=True
        )
        console_handler.setLevel(log_level)

        formatter = logging.Formatter(
            fmt="%(asctime)s - %(name)-15s - %(levelname)-8s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"
        )
        console_handler.setFormatter(formatter)

        self._root_logger.addHandler(console_handler)

        # Prevent propagation from the application logger into the Python root logger i.e.
        # the application logger is the final authority.
        self._root_logger.propagate = False


    def get_logger(self, component_name: str) -> PFBLogger:
        """
        Get a logger for a specific component.

        Args:
            component_name: Name of the component

        Returns:
            PFBLogger instance
        """
        if component_name not in self._loggers:
            self._loggers[component_name] = logging.getLogger(name=f"{self._app_name}.{component_name}")

        return self._loggers[component_name]

    def log_to_file(self, filename: Union[str, Path], component_name: Optional[str] = None) -> None:
        """
        Add file logging to the specified file.

        Args:
            filename: Path to the log file
            component_name: Optional component name to restrict file logging to specific component
        """
        filename = Path(filename)

        # Create directory if it doesn't exist
        filename.parent.mkdir(parents=True, exist_ok=True)

        # Create file handler
        file_handler = logging.FileHandler(filename, mode='a')
        file_handler.setLevel(logging.DEBUG)  # File gets all messages

        # File formatter (more detailed than console)
        file_formatter = logging.Formatter(
            fmt="%(asctime)s - %(name)-15s - %(levelname)-8s - %(funcName)s:%(lineno)d - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler.setFormatter(file_formatter)

        # Add to appropriate logger
        if component_name:  # Do we want to support this?
            # Add to specific component logger
            logger = self.get_logger(component_name)
            logger.logger.addHandler(file_handler)
        else:
            # Add to root logger (affects all components)
            self._root_logger.addHandler(file_handler)

        # Store the file reference
        key = component_name if component_name else 'root'
        self._log_files[key] = str(filename)

        # Log that we've started logging to file
        if component_name:
            logger = self.get_logger(component_name)
            logger.info(f"Logging to file: {filename}")
        else:
            # For root logger, log directly to avoid duplication
            # since child loggers will inherit the file handler
            self._root_logger.info(f"Logging to file: {filename}")

    def set_log_level(self, level: Union[str, int]) -> None:
        """
        Set the logging level for all loggers.

        Args:
            level: Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        """
        if isinstance(level, str):
            level = getattr(logging, level.upper())

        self._root_logger.setLevel(level)  # Should the application logger ever change level?

        # Update all handlers which are not instances of logging.FileHandler.
        for handler in self._root_logger.handlers:
            if isinstance(handler, logging.StreamHandler) and not isinstance(handler, logging.FileHandler):
                handler.setLevel(level)

    def get_log_files(self) -> Dict[str, str]:
        """Get dictionary of active log files."""
        return self._log_files.copy()

    def close_log_files(self) -> None:
        """Close all file handlers."""
        for handler in self._root_logger.handlers[:]:
            if isinstance(handler, logging.FileHandler):
                handler.close()
                self._root_logger.removeHandler(handler)

        # Clear file references
        self._log_files.clear()


# Global manager instance
_logging_manager = LoggingManager("pfb")


def get_logger(component_name: str) -> PFBLogger:
    """
    Get a logger for a specific component.

    Args:
        component_name: Name of the component

    Returns:
        PFBLogger instance
    """
    return _logging_manager.get_logger(component_name)


def log_to_file(filename: Union[str, Path], component_name: Optional[str] = None) -> None:
    """
    Add file logging to the specified file.

    Args:
        filename: Path to the log file
        component_name: Optional component name to restrict file logging to specific component
    """
    _logging_manager.log_to_file(filename, component_name)


def set_log_level(level: Union[str, int]) -> None:
    """
    Set the logging level for all loggers.

    Args:
        level: Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    """
    _logging_manager.set_log_level(level)


def get_log_files() -> Dict[str, str]:
    """Get dictionary of active log files."""
    return _logging_manager.get_log_files()


def close_log_files() -> None:
    """Close all file handlers."""
    _logging_manager.close_log_files()


# Utility functions for common logging patterns
def log_function_call(logger: PFBLogger):
    """
    Decorator to log function calls with arguments.

    Args:
        logger: PFBLogger instance to use for logging
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Log function entry
            logger.debug(f"Calling {func.__name__} with args={args}, kwargs={kwargs}")

            try:
                result = func(*args, **kwargs)
                logger.debug(f"Function {func.__name__} completed successfully")
                return result
            except Exception as e:
                logger.error(f"Function {func.__name__} failed with error: {e}")
                raise
        return wrapper
    return decorator


def log_options_dict(logger: PFBLogger, options: Dict[str, Any], title: str = "Options") -> None:
    """
    Log a dictionary of options in a formatted way.

    Args:
        logger: PFBLogger instance to use for logging
        options: Dictionary of options to log
        title: Title for the options section
    """

    name_col = Column(justify="left")
    spacer_col = Column(justify="center")
    val_col = Column(justify="left")

    options_table = Table.grid(name_col, spacer_col, val_col, expand=True)

    for key, value in options.items():
        options_table.add_row(f"[green]{key}[/green]", "|", f"[white]{value}[/white]")

    # We cannot log the table neatly, so instead we use console's print command directly
    # and capture the output so we can add an unformatted version to the log.
    with rich_console.capture() as capture:
        rich_console.print(Panel(options_table, style="cyan", title=f"[cyan]{title}[/cyan]"))

    str_output = Text.from_ansi(capture.get())
    rich_console.print(str_output.markup)  # Display in terminal.
    logger.debug(f"\n{str_output}")  # Print to log - will display twice if log level is DEBUG.



import numpy as np
import numexpr as ne
from pfb.utils.dist import l1reweight_func
from pfb.utils.misc import norm_diff
from numba import njit, prange
from uuid import uuid4
from pfb.utils import logging as pfb_logging
from time import time
log = pfb_logging.get_logger('PD')


def primal_dual(
        x,  # initial guess for primal variable
        v,  # initial guess for dual variable
        lam,  # regulariser strength
        psi,  # linear operator in dual domain
        psiH,  # adjoint of psi
        L,  # spectral norm of Hessian
        prox,  # prox of regulariser
        grad,  # gradient of smooth term
        nu=1.0,  # spectral norm of psi
        sigma=None,  # step size of dual update
        mask=None,  # regions where mask is False will be masked
        tol=1e-5,
        maxit=1000,
        minit=10,
        positivity=1,
        report_freq=10,
        gamma=1.0,
        verbosity=1):
    # initialise
    xp = x.copy()
    vp = v.copy()
    vtilde = np.zeros_like(v)
    vout = np.zeros_like(v)

    # this seems to give a good trade-off between
    # primal and dual problems
    if sigma is None:
        sigma = L / (2.0 * gamma) / nu

    # stepsize control
    tau = 0.9 / (L / (2.0 * gamma) + sigma * nu**2)

    # start iterations
    eps = 1.0
    k = 0
    while (eps > tol or k < minit) and k < maxit:
        # tmp prox variable
        vtilde = v + sigma * psiH(xp)

        # dual update
        v = vtilde - sigma * prox(vtilde / sigma, lam / sigma)

        # primal update
        x = xp - tau * (psi(2 * v - vp) + grad(xp))
        if positivity == 1:
            x[x < 0.0] = 0.0
        elif positivity == 2:
            msk = np.any(x<=0, axis=0)
            x[:, msk] = 0.0

        # convergence check
        eps = np.linalg.norm(x - xp) / np.linalg.norm(x)

        # copy contents to avoid allocating new memory
        xp[...] = x[...]
        vp[...] = v[...]

        if np.isnan(eps) or np.isinf(eps):
            import pdb; pdb.set_trace()

        if not k % report_freq and verbosity > 1:
            # res = xbar-x
            # phi = np.vdot(res, A(res))
            log.info(f"At iteration {k} eps = {eps:.3e}")
        k += 1

    if k == maxit:
        if verbosity:
            log.info(f"Max iters reached. eps = {eps:.3e}")
    else:
        if verbosity:
            log.info(f"Success, converged after {k} iterations")

    return x, v


from pfb.prox.prox_21m import dual_update_numba
def primal_dual_optimised(
        x,  # initial guess for primal variable
        v,  # initial guess for dual variable
        lam,  # regulariser strength
        psiH,  # linear operator in dual domain
        psi,  # adjoint of psi
        L,  # spectral norm of Hessian
        prox,  # prox of regulariser
        l1weight,
        reweighter,
        grad,  # gradient of smooth term
        nu=1.0,  # spectral norm of psi
        sigma=None,  # step size of dual update
        mask=None,  # regions where mask is False will be masked
        tol=1e-5,
        maxit=1000,
        positivity=1,
        report_freq=10,
        gamma=1.0,
        verbosity=1,
        maxreweight=20):  # max successive reweights before convergence

    # initialise
    xp = x.copy()
    vp = v.copy()
    xout = np.zeros_like(x)

    # this seems to give a good trade-off between
    # primal and dual problems
    if sigma is None:
        sigma = L / (2.0 * gamma) / nu

    # stepsize control
    tau = 0.98 / (L / (2.0 * gamma) + sigma * nu**2)

    # start iterations
    eps = 1.0
    numreweight = 0
    last_reweight_iter = 0
    # for timing
    tpsi = 0.0
    tupdate = 0.0
    teval1 = 0.0
    tpsiH = 0.0
    tgrad = 0.0
    teval2 = 0.0
    tpos = 0.0
    tnorm = 0.0
    tcopy = 0.0
    tii = time()
    for k in range(maxit):
        ti = time()
        psi(xp, v)
        tpsi += time() - ti
        ti = time()
        dual_update_numba(vp, v, lam,
                          sigma=sigma, weight=l1weight)
        tupdate += time() - ti
        ti = time()
        ne.evaluate('2.0 * v - vp', out=vp)  #, casting='same_kind')
        teval1 += time() - ti
        ti = time()
        psiH(vp, xout)
        tpsiH += time() - ti
        ti = time()
        xout += grad(xp)
        tgrad += time() - ti
        ti = time()
        ne.evaluate('xp - tau * xout', out=x)  #, casting='same_kind')
        teval2 += time() - ti

        ti = time()
        if positivity == 1:
            x[x < 0.0] = 0.0
        elif positivity == 2:
            msk = np.any(x<=0, axis=0)
            x[:, msk] = 0.0
        tpos += time() - ti
        # convergence check
        if x.any():
            ti = time()
            eps = norm_diff(x, xp)
            tnorm += time() - ti
        else:
            import pdb; pdb.set_trace()
            eps = 1.0
        if eps < tol:
            if reweighter is not None and numreweight < maxreweight:
                # ti = time()
                l1weight = reweighter(x)
                # log.info("reweight = ", time() - ti)
                if k-last_reweight_iter==1:
                    numreweight += 1
                else:
                    numreweight = 0
                last_reweight_iter = k

            else:
                if numreweight >= maxreweight:
                    log.info("Maximum reweighting steps reached")
                break

        # copy contents to avoid allocating new memory
        ti = time()
        np.copyto(xp, x)
        np.copyto(vp, v)
        tcopy += time() - ti
        if np.isnan(eps) or np.isinf(eps):
            import pdb; pdb.set_trace()

        if not k % report_freq and verbosity > 1:
            log.info(f"At iteration {k} eps = {eps:.3e}")

    ttot = time() - tii
    ttally = tpsi + tpsiH + tgrad + tupdate + teval1 + teval2 + tpos + tnorm
    if verbosity > 1:
        log.info('Time taken per step')
        log.info(f'psi = {tpsi/ttot}')
        log.info(f'psiH = {tpsiH/ttot}')
        log.info(f'grad = {tgrad/ttot}')
        log.info(f'update = {tupdate/ttot}')
        log.info(f'eval1 = {teval1/ttot}')
        log.info(f'eval2 = {teval2/ttot}')
        log.info(f'pos = {tpos/ttot}')
        log.info(f'norm = {tnorm/ttot}')
        log.info(f'tally = {ttally/ttot}')

    if k == maxit-1:
        if verbosity:
            log.info(f"Max iters reached. eps = {eps:.3e}")
    else:
        if verbosity:
            log.info(f"Success, converged after {k} iterations")

    return x, v

from distributed import as_completed

def primal_dual_dist(
            actors,
            lam,  # strength of regulariser
            L,    # spectral norm of Hessian
            l1weight,
            rmsfactor,
            rms_comps,
            alpha,
            nu=1.0,  # spectral norm of psi
            sigma=None,  # step size of dual update
            mask=None,  # regions where mask is False will be masked
            tol=1e-5,
            maxit=1000,
            positivity=1,
            report_freq=10,
            gamma=1.0,
            verbosity=1,
            maxreweight=50):
    '''
    Distributed primal dual algorithm.
    Distribution is over datasets in ddsf.

    Inputs:

    actors      - list of band actors
    lam         - strength of regulariser
    L           - spectral norm of hessian approximation
    l1weight    - array of L1 weights
    reweighter  - function to compute L1 reweights
    '''

    # this seems to give a good trade-off between
    # primal and dual problems
    if sigma is None:
        sigma = L / (2.0 * gamma) / nu

    # stepsize control
    tau = 0.9 / (L / (2.0 * gamma) + sigma * nu**2)

    # we need to do this upfront only at the outset
    futures = list(map(lambda a: a.init_pd_params(L, nu), actors))
    # we don't want to allocate at each iteration
    eps_num = [1.0]*len(futures)
    eps_den = [1.0]*len(futures)
    vtilde = np.zeros((len(actors), *l1weight.shape), dtype=l1weight.dtype)
    for fut in as_completed(futures):
        tmp, b = fut.result()
        vtilde[b] = tmp
    # vtilde = list(map(lambda f: f.result(), futures))
    numreweight = 0
    do_reweight = ~(l1weight == 1.0).all()  # reweighting active
    ratio = np.zeros(l1weight.shape, dtype=l1weight.dtype)
    for k in range(maxit):
        ti = time()
        get_ratio(vtilde, l1weight, sigma, lam, ratio)
        # get_ratio(np.array(vtilde), l1weight, sigma, lam, ratio)
        log.info('ratio - ', time() - ti)

        ti = time()
        # do on individual workers
        futures = list(map(lambda a: a.pd_update(ratio), actors))
        for fut in as_completed(futures):
            tmp, epsn, epsd, b = fut.result()
            vtilde[b] = tmp
            eps_num[b] = epsn
            eps_den[b] = epsd

        # results = list(map(lambda f: f.result(), futures))
        log.info('update - ', time() - ti)

        # vtilde = [r[0] for r in results]
        # eps_num = [r[1] for r in results]
        # eps_den = [r[2] for r in results]
        eps = np.sqrt(np.sum(eps_num)/np.sum(eps_den))

        if np.isnan(eps):
            log.error_and_raise("eps is nan",
                                ValueError)


        if not k % report_freq and verbosity > 1:
            log.info(f"At iteration {k} eps = {eps:.3e}")

        if eps < tol:
            if do_reweight and numreweight < maxreweight:
                l1weight = l1reweight_func(actors,
                                           rmsfactor,
                                           rms_comps=rms_comps,
                                           alpha=alpha)
                numreweight += 1
                log.info(f"Reweighting iter {numreweight}")
            else:
                if numreweight >= maxreweight:
                    log.info("Maximum reweighting steps reached")
                break

    if k >= maxit-1:
        log.info(f'Maximum iterations reached. eps={eps:.3e}')
    else:
        log.info(f'Success converged after {k} iterations')

    return


@njit(nogil=True, cache=True, parallel=True)
def get_ratio(vtilde, l1weight, sigma, lam, ratio):
    nband, nbasis, nymax, nxmax = vtilde.shape
    for b in range(nbasis):
        for i in prange(nymax):  # WTF without the prange it segfaults when parallel=True
            vtildebi = vtilde[:, b, i]
            weightbi = l1weight[b, i]
            for j in range(nxmax):
                vtildebij = vtildebi[:, j]
                weightbij = weightbi[j]
                absvbisum = np.abs(np.sum(vtildebij)/sigma)  # sum over band axis
                softvbisum = absvbisum - lam*weightbij/sigma
                if absvbisum > 0 and softvbisum > 0:
                    ratio[b, i, j] = softvbisum/absvbisum
                else:
                    ratio[b, i, j] = 0.0



primal_dual.__doc__ = r"""
    Algorithm to solve problems of the form

    argmin_x (xbar - x).T A (xbar - x)/2 + lam |PSI.H x|_21 s.t. x >= 0

    where x is the image cube and PSI an orthogonal basis for the spatial axes
    and the positivity constraint is optional.

    A        - Positive definite Hermitian operator
    xbar     - Data-like term
    x0       - Initial guess for primal variable
    v0       - Initial guess for dual variable
    lam      - Strength of l21 regulariser
    psi      - Orthogonal basis operator where
               psi.hdot() does decomposition and
               psi.dot() the reconstruction
               for all bases and channels
    weights  - Weights for l1 thresholding
    L        - Lipschitz constant of A
    nu       - Spectral norm of psi
    sigma    - l21 step size (set to L/2 by default)
    gamma    - step size during forward step

    Note that the primal variable (i.e. x) has shape (nband, nx, ny) where
    nband is the number of imaging bands and the dual variable (i.e. v) has
    shape (nbasis, nband, nmax) where nmax is the total number of coefficients
    for all bases. It is assumed that all bases are decomposed into the same
    number of coefficients. We use zero padding to deal with the fact that the
    basess do not necessarily have the same number of coefficients.
    """



import inspect
import numpy as np
import numba
import numexpr as ne
import xarray as xr
from pfb.opt.pcg import pcg
from pfb.operators.gridder import image_data_products, compute_residual
from pfb.operators.hessian import _hessian_slice
from pfb.operators.psf import psf_convolve_slice
from pfb.utils.weighting import _compute_counts
from pfb.utils.misc import taperf, set_image_size
from pfb.utils.naming import xds_from_url, xds_from_list
from uuid import uuid4
import pywt
from pfb.wavelets.wavelets_jsk import  (get_buffer_size,
                                       dwt2d, idwt2d)
from pfb.operators.psi import psi_band_maker
from ducc0.misc import empty_noncritical, resize_thread_pool
import concurrent.futures as cf
from africanus.constants import c as lightspeed
from ducc0.wgridder import vis2dirty, dirty2vis
from ducc0.fft import c2r, r2c, c2c, good_size
iFs = np.fft.ifftshift
Fs = np.fft.fftshift


class fake_future(object):
    '''
    This is just a class that has a fake result method.
    Useful for testing and when not running in distributed mode.
    '''
    def __init__(self, r):
        self.r = r

    def result(self):
        return self.r


class fake_client(object):
    '''
    This is just a class that has a fake submit method.
    Useful for testing and when not running in distributed mode.
    '''
    def __init__(self, nworkers=1, nvthreads=1):
        # TODO - use cf to fake these for horizontal parallelism
        self.nworkers = nworkers
        self.nvthreads = nvthreads

    def submit(self, *args, **kwargs):
        func = args[0]  # by convention
        params = inspect.signature(func).parameters
        fkw = {}
        for name, param in params.items():
            if name in kwargs:
                fkw[name] = kwargs[name]
        res = func(*args[1:], **fkw)
        # if 'actor' in kwargs and kwargs['actor']:  # return instantiated class
        #     return fake_actor(res)
        # else:  # return fake future with result
        return fake_future(res)

    def gather(self, futures):
        return [f.result() for f in futures]

    def close(self):
        return


def l1reweight_func(actors, rmsfactor, rms_comps=None, alpha=4):
    '''
    The logic here is that weights should remain the same for model
    components that are rmsfactor times larger than the rms of the updates.
    High SNR values should experience relatively small thresholding
    whereas small values should be strongly thresholded
    Set alpha > 1 for more agressive reweighting
    '''
    # project model
    futures = list(map(lambda a: a.psi_dot(mode='model'), actors))
    outvar = np.sum(list(map(lambda f: f.result(), futures)), axis=0)
    mcomps = np.abs(outvar)
    # we only project updates in the outer loop
    if rms_comps is None:
        futures = list(map(lambda a: a.psi_dot(mode='update'), actors))
        outvar = np.sum(list(map(lambda f: f.result(), futures)), axis=0)
        if not outvar.any():
            raise ValueError("Cannot reweight before any updates have been performed")
        # exclude zero components due to padding wavelets
        rms_comps = np.std(outvar[outvar!=0])
        return (1 + rmsfactor)/(1 + mcomps**alpha/rms_comps**alpha), rms_comps
    else:
        assert rms_comps > 0
        return (1 + rmsfactor)/(1 + mcomps**alpha/rms_comps**alpha)


class band_actor(object):
    def __init__(self, xds_list, opts, bandid, cache_path, max_freq, uv_max):
        self.opts = opts
        self.bandid = bandid
        self.cache_path = f'{cache_path}/time0000_band{bandid:04d}.zarr'
        self.xds_list = xds_list
        self.nthreads = opts.nthreads
        self.real_type = np.float64
        self.complex_type = np.complex128
        # hess_approx determines whether we use the vis of image space
        # approximation of the hessian.
        self.hess_approx = opts.hess_approx
        resize_thread_pool(self.nthreads)
        numba.set_num_threads(self.nthreads)

        dsl = xds_from_list(xds_list, drop_vars=['VIS', 'BEAM'],
                            nthreads=self.nthreads)

        times_in = []
        self.freq = dsl[0].FREQ.values
        for ds in dsl:
            times_in.append(ds.time_out)
            if (ds.FREQ.values != self.freq).all():
                raise NotImplementedError("Freqs per band currently assumed to be the same")

        times_in = np.unique(times_in)

        # drop before concat
        for i, ds in enumerate(dsl):
            ds = ds.drop_vars('FREQ')
            dsl[i] = ds.drop_dims(('l_beam', 'm_beam'))

        # concat and check if band is empty
        xds = xr.concat(dsl, dim='row')
        wsumb = np.sum(xds.WEIGHT.values*xds.MASK.values)
        if wsumb == 0:
            raise ValueError('Got empty dataset. This is a bug!')

        # set cell sizes
        self.max_freq = max_freq
        self.uv_max = uv_max
        nx, ny, nx_psf, ny_psf, cell_N, cell_rad = set_image_size(uv_max,
                                                                  max_freq,
                                                                  opts.field_of_view,
                                                                  opts.super_resolution_factor,
                                                                  opts.cell_size,
                                                                  opts.nx,
                                                                  opts.ny,
                                                                  opts.psf_oversize)
        cell_deg = np.rad2deg(cell_rad)
        cell_size = cell_deg * 3600
        # log.info(f"Super resolution factor = {cell_N/cell_rad}")
        # log.info(f"Cell size set to {cell_size:.5e} arcseconds")
        # log.info(f"Field of view is ({nx*cell_deg:.3e},{ny*cell_deg:.3e}) degrees")

        self.cell_rad = cell_rad
        self.nx = nx
        self.ny = ny
        self.nx_psf = nx_psf
        self.ny_psf = ny_psf
        self.nyo2 = self.ny_psf//2 + 1
        if nx_psf > nx or ny_psf > ny:
            npad_xl = (nx_psf - nx) // 2
            npad_xr = nx_psf - nx - npad_xl
            npad_yl = (ny_psf - ny) // 2
            npad_yr = ny_psf - ny - npad_yl
            self.padding = ((npad_xl, npad_xr), (npad_yl, npad_yr))
            self.unpad_x = slice(npad_xl, -npad_xr)
            self.unpad_y = slice(npad_yl, -npad_yr)
        else:
            self.padding = ((0, 0), (0, 0))
            self.unpad_x = slice(None)
            self.unpad_y = slice(None)

        self.ra = dsl[0].ra
        self.dec = dsl[0].dec
        self.x0 = 0.0
        self.y0 = 0.0

        self.freq_out = np.mean(self.freq)
        self.time_out = np.mean(times_in)

        self.counts = _compute_counts(
                                xds.UVW.values,
                                self.freq,
                                xds.MASK.values,
                                xds.WEIGHT.values,
                                nx, ny,
                                cell_rad, cell_rad,
                                self.real_type,
                                ngrid=self.nthreads)

        # set up wavelet dictionaries
        self.bases = opts.bases.split(',')
        self.nbasis = len(self.bases)
        self.nlevel = opts.nlevels
        self.psi = psi_band_maker(self.nx, self.ny, self.bases, self.nlevel)
        self.Nxmax = self.psi.Nxmax
        self.Nymax = self.psi.Nymax

        # tmp optimisation vars
        self.b = np.random.randn(self.nx, self.ny).astype(self.real_type)
        self.bp = np.zeros((self.nx, self.ny), dtype=self.real_type)
        self.xp = np.zeros((self.nx, self.ny), dtype=self.real_type)
        self.vp = np.zeros((self.nbasis, self.Nymax, self.Nxmax), dtype=self.real_type)
        self.vtilde = np.zeros((self.nbasis, self.Nymax, self.Nxmax), dtype=self.real_type)

        # always initialised to zero
        self.dual = np.zeros((self.nbasis, self.Nymax, self.Nxmax), dtype=self.real_type)
        self.dual_tmp = np.zeros((self.nbasis, self.Nymax, self.Nxmax), dtype=self.real_type)

        # this will be overwritten if loading from cache
        # important for L1-reweighting!
        self.update = np.zeros((self.nx, self.ny), dtype=self.real_type)

        # pre-allocate tmp arrays required for FFTs
        # make_noncritical for efficiency of FFT
        self.xhat = empty_noncritical((self.nx_psf, self.nyo2),
                                       dtype=self.complex_type)
        self.xpad = empty_noncritical((self.nx_psf, self.ny_psf),
                                       dtype=self.real_type)
        self.xout = empty_noncritical((self.nx, self.ny),
                                       dtype=self.real_type)


        self.eta = opts.eta
        if opts.hess_approx=='wgt':
            self.hess = self.hess_wgt
        elif opts.hess_approx=='psf':
            self.hess = self.hess_psf
        elif opts.hess_approx=='direct':
            self.hess = self.hess_direct
            if self.eta < 1:
                log.info(f'Warning - using dangerously low value of '
                      f'{self.eta} with direct Hessian approximation')
        npix = np.maximum(self.nx, self.ny)
        self.taperxy = taperf((self.nx, self.ny), int(0.1*npix))

        self.attrs = {
            'ra': self.ra,
            'dec': self.dec,
            'x0': self.x0,
            'y0': self.y0,
            'cell_rad': self.cell_rad,
            'bandid': self.bandid,
            'timeid': '0000',
            'freq_out': self.freq_out,
            'time_out': self.time_out,
            'robustness': self.opts.robustness,
            'super_resolution_factor': self.opts.super_resolution_factor,
            'field_of_view': self.opts.field_of_view,
            'product': self.opts.product
        }

        # check that the numba threading layer is available
        self.psi.dot(self.b, self.vp)
        self.psi.hdot(self.vp, self.bp)
        assert (self.nbasis*self.b - self.bp < 1e-12).all()

        try:
            tlayer = numba.threading_layer()
            assert tlayer == 'tbb'
        except Exception as e:
            log.info('Warning - numba threading layer not initialised correctly')
            log.info('Original traceback', e)


    def get_image_info(self):
        return (self.nx, self.ny, self.Nymax, self.Nxmax, self.cell_rad,
                self.ra, self.dec, self.x0, self.y0,
                self.freq_out, self.time_out)


    def set_image_data_products(self, model, k, dof=None, from_cache=False):
        '''
        This initialises the psf and residual.
        Also used to perform L2 reweighting so wsum can change.
        This is why we can't normalise by wsum.
        Returns the per-band resid and wsum so that we can
        compute the MFS images (and hence wsum, rms and rmax etc.)
        on the runner.
        '''

        self.model = model

        if from_cache:
            self.dds = xr.open_zarr(self.cache_path, chunks=None)
            self.update = self.dds.UPDATE.values
            residual = self.dds.RESIDUAL.values
            self.wsumb = self.dds.wsum
        else:
            # attempt to save memory
            if hasattr(self, 'dds'):
                self.dds = None
            residual, wsumb = image_data_products(
                                self.xds_list,
                                self.counts,
                                self.nx, self.ny,
                                self.nx_psf, self.ny_psf,
                                self.cell_rad, self.cell_rad,
                                self.cache_path,
                                self.attrs,
                                model=model,
                                robustness=self.opts.robustness,
                                x0=self.x0, y0=self.y0,
                                nthreads=self.nthreads,
                                epsilon=self.opts.epsilon,
                                do_wgridding=self.opts.do_wgridding,
                                double_accum=self.opts.double_accum,
                                # divide_by_n=False,
                                l2_reweight_dof=dof,
                                do_dirty=True,
                                do_psf=True,
                                do_residual=True,
                                do_weight=True,
                                do_noise=False)

            self.wsumb = wsumb
            # cache vars affected by major cycle
            # we cache the weights in the
            data_vars = {
                'UPDATE': (('x','y'), self.update)
            }
            self.attrs['niters'] = k
            self.attrs['wsum'] = wsumb
            dds = xr.Dataset(data_vars, attrs=self.attrs)
            dds.to_zarr(self.cache_path, mode='a')

            # these are read in each major cycle anyway
            drop_vars = ['PSF','MODEL','DIRTY']
            if self.opts.hess_approx != 'wgt':
                drop_vars += ['WEIGHT','UVW','MASK']
            self.dds = xds_from_list([self.cache_path],
                                     drop_vars=drop_vars)[0]

        # return the per band resid since resid MFS required on runner
        return residual, self.wsumb

    def set_wsum(self, wsum):
        self.wsum = wsum  # wsum over all bands
        # we can only normalise once we have wsum from all bands
        # only required if wsum changes
        self.psfhat = self.dds.PSFHAT.values/wsum
        self.residual = self.dds.RESIDUAL.values/wsum
        if self.opts.hess_approx == 'direct':
            self.psfhat = np.abs(self.psfhat)
            # # Briggs weighting factor
            # robust = 0.5
            # numsqrt = 5*10**(-robust)
            # avgW = (self.psfhat ** 2).sum() / self.psfhat.sum()
            # ssq = numsqrt * numsqrt/avgW
            # self.taper = 1 + self.psfhat * ssq


    def set_residual(self, k, x=None):
        if x is None:
            x = self.model

        if self.wsumb == 0:
            self.data = np.zeros_like(x)
            return np.zeros_like(x)

        # we don;t need counts here because we use the weights in the dds
        residual, _ = compute_residual(
                                self.cache_path,
                                self.nx, self.ny,
                                self.cell_rad, self.cell_rad,
                                self.cache_path,  # output_name (same as dsl names?)
                                x,
                                nthreads=self.nthreads,
                                epsilon=self.opts.epsilon,
                                do_wgridding=self.opts.do_wgridding,
                                double_accum=self.opts.double_accum)

        # is this even necessary?
        data_vars = {
            'UPDATE': (('x','y'), self.update)
        }
        self.attrs['niters'] = k
        self.attrs['wsum'] = self.wsumb
        dds = xr.Dataset(data_vars, attrs=self.attrs)
        dds.to_zarr(self.cache_path, mode='a')

        # these are read in each major cycle anyway
        drop_vars = ['PSF','MODEL','DIRTY']
        if self.opts.hess_approx != 'wgt':
            drop_vars += ['WEIGHT','UVW','MASK']
        self.dds = xds_from_list([self.cache_path],
                                    drop_vars=drop_vars)[0]

        # wsum doesn't change so we can normalise
        self.residual = residual/self.wsum

        # return residual since we need the MFS residual on the runner
        return residual

    def hess_direct(self, x, mode='forward', sigma=None):
        if sigma is None:
            sigma = self.eta
        if self.wsumb == 0:  # should not happen
            return np.zeros((self.nx, self.ny), dtype=self.real_type)
        self.xpad[...] = 0.0
        self.xpad[self.unpad_x, self.unpad_y] = x*self.taperxy
        r2c(self.xpad, out=self.xhat,
            forward=True, inorm=0, nthreads=self.nthreads)
        if mode=='forward':
            self.xhat *= (self.psfhat + sigma)
        elif mode=='backward':
            self.xhat /= (self.psfhat + sigma)
        c2r(self.xhat, out=self.xpad, lastsize=self.ny_psf,
            forward=False, inorm=2, nthreads=self.nthreads,
            allow_overwriting_input=True)
        self.xout[...] = self.xpad[self.unpad_x, self.unpad_y] * self.taperxy
        return self.xout

    def hess_psf(self, x, mode='forward'):
        raise NotImplementedError('Not yet')


    def hess_wgt(self, x, mode='forward'):
        convx = np.zeros_like(x)
        if self.wsumb == 0:
            return convx

        if mode=='forward':
            convx += _hessian_slice(
                                x,
                                self.dds.UVW.values,
                                self.dds.WEIGHT.values,
                                self.dds.MASK.values,
                                self.freq,
                                None, # self.beam
                                x0=0.0, y0=0.0,
                                cell=self.cell_rad,
                                do_wgridding=self.opts.do_wgridding,
                                epsilon=5e-4,  # we don't need much accuracy here
                                double_accum=self.opts.double_accum,
                                nthreads=self.nthreads)

            return convx/self.wsum + self.eta * x
        else:
            raise NotImplementedError('Not yet')

    def cg_update(self):
        if self.hess_approx == 'wgt':
            # precond = lambda x: self.hess_direct(x,
            #                                      mode='backward',
            #                                      sigma=1.0)
            x = pcg(
                A=self.hess_wgt,
                b=self.residual,
                # x0=precond(self.residual),
                # M=precond,
                tol=self.opts.cg_tol,
                maxit=self.opts.cg_maxit,
                minit=self.opts.cg_minit,
                verbosity=self.opts.cg_verbose,
                report_freq=self.opts.cg_report_freq,
                backtrack=self.opts.backtrack,
                return_resid=False
            )
        elif self.hess_approx == "psf":
            # the information source differs because we need a Hermitian
            # postive definite operator for the pcg

            raise NotImplementedError
        elif self.hess_approx == "direct":
            x = self.hess_direct(self.residual, mode='backward')
        else:
            raise ValueError(f'Unknown hess-approx {self.hess_approx}')

        self.update = x
        self.xtilde = self.model + x

        return x

    def backward_grad(self, x):
        '''
        The gradient function during the backward step
        '''
        return -self.hess(self.xtilde - x, mode='forward')

    def psi_dot(self, x=None, mode='model'):
        # TODO - this is a bit of a kludge that we need
        # for performing L1 reweights
        if mode == 'model':
            self.psi.dot(self.model, self.dual)
            return self.dual
        elif mode == 'update':
            # we don't want to bias rms_comps due to tapered edges
            self.psi.dot(self.update/self.taperxy, self.dual)
            return self.bp

    def init_pd_params(self, hess_norm, nu, gamma=1):
        self.hess_norm = hess_norm
        self.sigma = hess_norm/(2*gamma*nu)
        self.tau = 0.9 / (hess_norm / (2.0 * gamma) + self.sigma * nu**2)
        self.psi.dot(self.model, self.vtilde)
        self.vtilde *= self.sigma
        self.vtilde += self.dual
        # self.vtilde = self.dual + self.sigma * self.psi_dot(x=self.model)
        return self.vtilde, int(self.bandid)

    def pd_update(self, ratio):
        # ratio - (nbasis, nymax, nxmax)
        # ti = time.time()
        self.xp[...] = self.model[...]
        self.vp[...] = self.dual[...]
        # log.info('copy = ', time.time() - ti)
        self.dual[...] = self.vtilde * (1 - ratio)

        # ti = time.time()
        grad = self.backward_grad(self.xp)
        # log.info('grad = ', time.time() - ti)
        # ti = time.time()
        self.psi.hdot(2*self.dual - self.vp, self.b)
        self.model[...] = self.xp - self.tau * (self.b + grad)
        # log.info('update = ', time.time() - ti)

        # ti = time.time()
        if self.opts.positivity:
            self.model[self.model < 0] = 0.0
        # log.info('postivity = ', time.time() - ti)

        # ti = time.time()
        # do in place
        self.psi.dot(self.model, self.vtilde)
        self.vtilde *= self.sigma
        self.vtilde += self.dual
        # self.vtilde[...] = self.dual + self.sigma * self.psi_dot(x=self.model)
        # log.info('vtilde = ', time.time() - ti)

        # ti = time.time()
        eps_num = np.sum((self.model-self.xp)**2)
        eps_den = np.sum(self.model**2)
        # log.info('eps = ', time.time() - ti)

        # vtilde - (nbasis, nymax, nxmax)
        return self.vtilde, eps_num, eps_den, int(self.bandid)

    # # the ne version doesn't seem to make a huge difference
    # def pd_update_ne(self, ratio):
    #     # ratio - (nbasis, nymax, nxmax)
    #     # ti = time.time()
    #     self.xp[...] = self.model[...]
    #     self.vp[...] = self.dual[...]
    #     # log.info('copy = ', time.time() - ti)
    #     # self.dual[...] = self.vtilde * (1 - ratio)
    #     ne.evaluate('vtilde * (1 - ratio)', local_dict={
    #         'vtilde' : self.vtilde,
    #         'ratio': ratio},
    #         out=self.dual, casting='same_kind')

    #     # ti = time.time()
    #     grad = self.backward_grad(self.xp)
    #     # log.info('grad = ', time.time() - ti)
    #     # ti = time.time()
    #     ne.evaluate('2*dual - vp', local_dict={
    #         'dual': self.dual,
    #         'vp': self.vp},
    #         out=self.dual_tmp, casting='same_kind')
    #     # self.psi.hdot(2*self.dual - self.vp, self.b)
    #     self.psi.hdot(self.dual_tmp, self.b)
    #     ne.evaluate('xp - tau*(b + g)', local_dict={
    #         'xp': self.xp,
    #         'tau': self.tau,
    #         'b': self.b,
    #         'g': grad},
    #         out=self.model, casting='same_kind')
    #     # self.model[...] = self.xp - self.tau * (self.b + grad)
    #     # log.info('update = ', time.time() - ti)

    #     # ti = time.time()
    #     if self.opts.positivity:
    #         self.model[self.model < 0] = 0.0
    #     # log.info('postivity = ', time.time() - ti)

    #     # ti = time.time()
    #     self.psi.dot(self.model, self.vtilde)
    #     ne.evaluate('d + sigma * vtilde', local_dict={
    #         'd': self.dual,
    #         'sigma': self.sigma,
    #         'vtilde': self.vtilde},
    #         out=self.vtilde, casting='same_kind')
    #     # self.vtilde *= self.sigma
    #     # self.vtilde += self.dual
    #     # self.vtilde[...] = self.dual + self.sigma * self.psi_dot(x=self.model)
    #     # log.info('vtilde = ', time.time() - ti)

    #     # ti = time.time()
    #     eps_num = ne.evaluate('sum((m-xp)**2)', local_dict={
    #         'm': self.model,
    #         'xp': self.xp},
    #         casting='same_kind')
    #     # eps_num = np.sum((self.model-self.xp)**2)
    #     eps_den = np.sum(self.model**2)
    #     # log.info('eps = ', time.time() - ti)

    #     # vtilde - (nband, nbasis, nymax, nxmax)
    #     return self.vtilde, eps_num, eps_den, int(self.bandid)

    def init_random(self):
        self.b[...] = np.random.randn(self.nx, self.ny)
        return np.sum(self.b**2)


    def pm_update(self, bnorm):
        self.bp[...] = self.b/bnorm
        self.b[...] = self.hess(self.bp)
        bsumsq = np.sum(self.b**2)
        beta_num = np.vdot(self.b, self.bp)
        beta_den = np.vdot(self.bp, self.bp)

        return bsumsq, beta_num, beta_den

    def give_model(self):
        return self.model, self.bandid



import ray
import concurrent.futures as cf
import numpy as np
from numba import njit, prange, literally, types
from numba.extending import overload
from pfb.utils.misc import JIT_OPTIONS, _es_kernel
from pfb.utils.stokes import stokes_funcs
from pfb.utils.naming import xds_from_list
from scipy.constants import c as lightspeed
iFs = np.fft.ifftshift
Fs = np.fft.fftshift

JIT_OPTIONS = {
    "nogil": True,
    "cache": True,
    "error_model": 'numpy',
    # "fastmath": True  # causes weighting tests to fail
}

@njit(**JIT_OPTIONS, inline='always')
def _es_kernel(x, y, xkern, ykern, betak):
    for i in range(x.size):
        if x[i]*x[i] < 1:
            xkern[i] = np.exp(betak*(np.sqrt(1-x[i]*x[i]) - 1))
        else:
            xkern[i] = 0.0
        if y[i]*y[i] < 1:
            ykern[i] = np.exp(betak*(np.sqrt(1-y[i]*y[i]) - 1))
        else:
            ykern[i] = 0.0

def compute_counts(dsl,
                   nx, ny,
                   cell_size_x, cell_size_y,
                   tbid=0,
                   k=6,
                   nthreads=1):
    '''
    Sum the weights on the grid over all datasets
    '''

    if isinstance(dsl, str):
        dsl = [dsl]

    dsl = xds_from_list(dsl, nthreads=nthreads,
                        drop_vars=('VIS','BEAM'))

    nds = len(dsl)
    maxw = np.minimum(nthreads, nds)
    if nthreads > nds:  # this is not perfect
        ngrid = np.maximum(nthreads//nds, 1)
    else:
        ngrid = 1

    counts = np.zeros((nx, ny), dtype=dsl[0].WEIGHT.dtype)
    with cf.ThreadPoolExecutor(max_workers=maxw) as executor:
        futures = []
        for ds in dsl:
            fut = executor.submit(_compute_counts,
                                  ds.UVW.values,
                                  ds.FREQ.values,
                                  ds.MASK.values,
                                  ds.WEIGHT.values,
                                  nx, ny,
                                  cell_size_x,
                                  cell_size_y,
                                  ds.WEIGHT.dtype,
                                  k=6,
                                  ngrid=ngrid)
            futures.append(fut)

        for fut in cf.as_completed(futures):
            # sum over number of datasets
            counts += fut.result().sum(axis=0)

    return counts, tbid



@njit(nogil=True, cache=True, parallel=True)
def _compute_counts(uvw, freq, mask, wgt, nx, ny,
                    cell_size_x, cell_size_y, dtype,
                    k=6, ngrid=1, usign=1.0, vsign=-1.0):  # support hardcoded for now
    # ufreq
    u_cell = 1/(nx*cell_size_x)
    # factor of 2 because -umax/2 <= u < umax
    umax = np.abs(1/cell_size_x/2)

    # vfreq
    v_cell = 1/(ny*cell_size_y)
    vmax = np.abs(1/cell_size_y/2)

    ncorr, nrow, nchan = wgt.shape

    # initialise array to store counts
    # the additional axis is to allow chunking over row
    counts = np.zeros((ngrid, ncorr, nx, ny), dtype=dtype)

    # accumulate counts
    bin_counts = [nrow // ngrid + (1 if x < nrow % ngrid else 0)  for x in range (ngrid)]
    bin_idx = np.zeros(ngrid, dtype=np.int64)
    bin_counts = np.asarray(bin_counts).astype(bin_idx.dtype)
    bin_idx[1:] = np.cumsum(bin_counts)[0:-1]

    ko2 = k//2
    betak = 2.3*k
    pos = np.arange(k) - ko2
    xkern = np.zeros(k, dtype=dtype)
    ykern = np.zeros(k, dtype=dtype)
    x = np.zeros(k, dtype=dtype)
    y = np.zeros(k, dtype=dtype)
    x_idx = np.zeros(k, dtype=np.int64)
    y_idx = np.zeros(k, dtype=np.int64)
    for g in prange(ngrid):
        for r in range(bin_idx[g], bin_idx[g] + bin_counts[g]):
            uvw_row = uvw[r]
            wgt_row = wgt[:, r]
            mask_row = mask[r]
            for f in range(nchan):
                if not mask_row[f]:
                    continue
                # current uv coords
                chan_normfreq = freq[f] / lightspeed
                u_tmp = uvw_row[0] * chan_normfreq * usign
                v_tmp = uvw_row[1] * chan_normfreq * vsign
                if v_tmp < 0:
                    u_tmp = -u_tmp
                    v_tmp = -v_tmp
                # pixel coordinates
                ug = (u_tmp + umax)/u_cell
                vg = (v_tmp + vmax)/v_cell
                # indices
                u_idx = np.int32(np.floor(ug))
                v_idx = np.int32(np.floor(vg))

                # corr weights
                wrf = wgt_row[:, f]
                
                # LB - is there an easier check for this?
                if (u_idx<0) or (u_idx>nx) or (v_idx<0) or (v_idx>ny):
                    print('uv out of bounds in cc')
                    continue

                # nearest neighbour
                if k==0:
                    counts[g, :, u_idx, v_idx] += wrf
                    continue

                # the kernel is separable and only defined on [-1,1]
                x_idx[:] = pos + u_idx
                x[:] = (x_idx - ug + 0.5)/ko2
                y_idx[:] = pos + v_idx
                y[:] = (y_idx - vg + 0.5)/ko2
                _es_kernel(x, y, xkern, ykern, betak)
                # check bounds
                valid_ix = np.nonzero((x_idx >= 0) & (x_idx < nx))[0]
                valid_iy = np.nonzero((y_idx >= 0) & (y_idx < ny))[0]

                for c in range(ncorr):
                    wrfc = wrf[c]
                    for i in valid_ix:
                        ix = x_idx[i]
                        xi = xkern[i]
                        for j in valid_iy:
                            iy = y_idx[j]
                            yi = ykern[j]
                            counts[g, c, ix, iy] += xi*yi*wrfc

    return counts.sum(axis=0)


@njit(**JIT_OPTIONS)
def counts_to_weights(counts, uvw, freq, weight, mask, nx, ny,
                      cell_size_x, cell_size_y, robust,
                      usign=1.0, vsign=-1.0):
    # when does this happen?
    if not counts.any():
        return weight

    real_type = weight.dtype

    # ufreq
    u_cell = 1/(nx*cell_size_x)
    umax = np.abs(1/cell_size_x/2)

    # vfreq
    v_cell = 1/(ny*cell_size_y)
    vmax = np.abs(1/cell_size_y/2)

    ncorr, nrow, nchan = weight.shape

    # Briggs weighting factor
    if robust > -2:
        numsqrt = 5*10**(-robust)
        avgWnum = np.zeros(ncorr, dtype=real_type)
        avgWden = np.zeros(ncorr, dtype=real_type)
        for c in range(ncorr):
            for i in range(nx):
                for j in range(ny):
                    cval = counts[c, i, j]
                    avgWnum[c] += cval*cval
                    avgWden[c] += cval
        ssq = numsqrt * numsqrt * avgWden / avgWnum
        counts *= ssq[:, None, None]
        counts += 1

    for r in prange(nrow):
        uvw_row = uvw[r]
        wgt_row = weight[:, r]
        mask_row = mask[r]
        for f in range(nchan):
            if not mask_row[f]:
                continue
            # current uv coords
            chan_normfreq = freq[f] / lightspeed
            u_tmp = uvw_row[0] * chan_normfreq * usign
            v_tmp = uvw_row[1] * chan_normfreq * vsign
            if v_tmp < 0:
                u_tmp = -u_tmp
                v_tmp = -v_tmp
            # pixel coordinates
            ug = (u_tmp + umax)/u_cell
            vg = (v_tmp + vmax)/v_cell

            # indices
            u_idx = np.int32(np.floor(ug))
            v_idx = np.int32(np.floor(vg))

            if (u_idx<0) or (u_idx>nx) or (v_idx<0) or (v_idx>ny):
                print('uv is out of bounds in c2w')
                continue

            # counts can be zero if there are zero weights
            if not np.any(counts[:, u_idx, v_idx] == 0):
                wgt_row[:, f] /= counts[:, u_idx, v_idx]

    return weight

# @njit(nogil=True, cache=True)
def filter_extreme_counts(counts, level=10.0):
    '''
    Replaces extremely small counts by median to prevent
    upweighting nearly empty cells
    '''
    # get the median counts value
    if not level:
        return counts
    ic, ix, iy = np.where(counts > 0)
    cnts = counts[ic, ix,iy]
    med = np.median(cnts)
    lowval = med/level
    cnts = np.maximum(cnts, lowval)
    counts[ic, ix,iy] = cnts
    return counts


@njit(nogil=True, cache=False, parallel=False)
def weight_data(data, weight, flag, jones, tbin_idx, tbin_counts,
                ant1, ant2, pol, product, nc):

    vis, wgt = _weight_data_impl(data, weight, flag, jones,
                                 tbin_idx, tbin_counts,
                                 ant1, ant2,
                                 literally(pol),
                                 literally(product),
                                 literally(nc))

    return vis, wgt


def _weight_data_impl(data, weight, flag, jones, tbin_idx, tbin_counts,
                 ant1, ant2, pol, product, nc):
    raise NotImplementedError


@overload(_weight_data_impl, prefer_literal=True,
          jit_options={**JIT_OPTIONS, "parallel":True})
def nb_weight_data_impl(data, weight, flag, jones, tbin_idx, tbin_counts,
                      ant1, ant2, pol, product, nc):

    vis_func, wgt_func = stokes_funcs(data, jones, product, pol, nc)
    ns = len(product.literal_value)

    def _impl(data, weight, flag, jones, tbin_idx, tbin_counts,
              ant1, ant2, pol, product, nc):
        # for dask arrays we need to adjust the chunks to
        # start counting from zero
        tbin_idx = tbin_idx - tbin_idx.min()
        nt = np.shape(tbin_idx)[0]
        nrow, nchan, ncorr = data.shape
        vis = np.zeros((nrow, nchan, ns), dtype=data.dtype)
        wgt = np.zeros((nrow, nchan, ns), dtype=data.real.dtype)

        for t in prange(nt):
            for row in range(tbin_idx[t],
                             tbin_idx[t] + tbin_counts[t]):
                p = int(ant1[row])
                q = int(ant2[row])
                gp = jones[t, p, :, 0]
                gq = jones[t, q, :, 0]
                for chan in range(nchan):
                    if flag[row, chan].any():
                        continue
                    wgt[row, chan] = wgt_func(gp[chan], gq[chan],
                                              weight[row, chan])
                    vis[row, chan] = vis_func(gp[chan], gq[chan],
                                              weight[row, chan],
                                              data[row, chan])

        return (vis, wgt)

    _impl.returns = types.Tuple([types.Array(types.complex128, 3, 'C'),
                                 types.Array(types.float64, 3, 'C')])

    return _impl


@njit
def weight_data_np(data, weight, flag, jones, tbin_idx, tbin_counts, ant1, ant2, ns, vis_func, wgt_func):
    # for dask arrays we need to adjust the chunks to
    # start counting from zero
    tbin_idx = tbin_idx - tbin_idx.min()
    nt = np.shape(tbin_idx)[0]
    nrow, nchan, ncorr = data.shape
    vis = np.zeros((nrow, nchan, ns), dtype=data.dtype)
    wgt = np.zeros((nrow, nchan, ns), dtype=data.real.dtype)

    for t in prange(nt):
        for row in range(tbin_idx[t],
                            tbin_idx[t] + tbin_counts[t]):
            p = int(ant1[row])
            q = int(ant2[row])
            gp = jones[t, p, :, 0]
            gq = jones[t, q, :, 0]
            for chan in range(nchan):
                if flag[row, chan].any():
                    continue
                wgt[row, chan] = wgt_func(gp[chan], gq[chan],
                                            weight[row, chan])
                vis[row, chan] = vis_func(gp[chan], gq[chan],
                                            weight[row, chan],
                                            data[row, chan])

    return (vis, wgt)


import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
import time
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('SMOOVIE')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.smoovie)
def smoovie(**kw):
    '''
    Smooth high cadence imaging results
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    ncpu = psutil.cpu_count(logical=False)
    # to prevent flickering
    opts.nthreads = 1
    remprod = opts.product.upper().strip('IQUV')
    if len(remprod):
        log.error_and_raise(f"Product {remprod} not yet supported",
                            NotImplementedError)

    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/smoovie_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')
    OmegaConf.set_struct(opts, True)

    pfb_logging.log_options_dict(log, opts)

    from pfb import set_envs
    set_envs(opts.nthreads, ncpu)

    # import dask
    # dask.config.set(**{'array.slicing.split_large_chunks': False})
    # from pfb import set_client
    # client = set_client(opts.nworkers, log, client_log_level=opts.log_level)

    ti = time.time()
    _smoovie(**opts)

    log.info(f"All done after {time.time() - ti}s")

    try:
        from distributed import get_client
        client = get_client()
        client.close()
    except Exception as e:
        pass


def _smoovie(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)


    import xarray as xr
    import numpy as np
    from pfb.utils.naming import xds_from_url
    import matplotlib.pyplot as plt
    from streamjoy import stream, wrap_matplotlib
    from daskms.fsspec_store import DaskMSStore

    basename = opts.output_filename
    if opts.scratch_dir is not None:
        scratch_dir = opts.scratch_dir
    else:
        scratch_dir = basename.rsplit('/', 1)[0]

    # xds contains vis products, no imaging weights applied
    fds_name = f'{basename}.fds' if opts.fds is None else opts.fds
    fds_store = DaskMSStore(fds_name)
    try:
        assert fds_store.exists()
    except Exception as e:
        log.error_and_raise(f"There must be a dataset at {fds_store.url}",
                            ValueError)

    log.info(f"Lazy loading fds from {fds_store.url}")
    fds, fds_list = xds_from_url(fds_store.url)

    # TODO - scan selection

    # get input times and frequencies
    freqs_in = []
    times_in = []
    for ds in fds:
        freqs_in.append(ds.FREQ.values)
        times_in.append(ds.TIME.values)

    freqs_in = np.unique(np.array(freqs_in))
    times_in = np.unique(np.array(times_in))

    @wrap_matplotlib()
    def plot_frame(ds):
        wsum = ds.wsum
        utc = ds.utc
        scan = ds.scanid
        fnum = ds.ffrac
        band = ds.bandid
        resid = ds.cube.values[0, 0, :, :]
        # nstokes = resid.shape[0]
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 10))
        fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
        # for i, ax in enumerate(axs):
        rms = ds.rms[0, 0]
        im1 = ax.imshow(resid,
                    vmin=-opts.min_frac*rms,
                    vmax=opts.max_frac*rms,
                    cmap=opts.cmap)

        plt.xticks([]), plt.yticks([])
        ax.annotate(
            f'{basename}_band{band:04d}_scan{scan:04d}_{opts.product[0]}' + '\n' + fnum + '\n' + utc,
            xy=(0.0, 0.0),
            xytext=(0.05, 0.05),
            xycoords='axes fraction',
            textcoords='axes fraction',
            ha='left', va='bottom',
            fontsize=15,
            color=opts.text_colour)
        return fig

    outfmt = opts.out_format.lower()
    idfy = f'fps{opts.fps}'
    if opts.animate_axis == 'time':
        # bin freq axis and make movie for each bin
        fds_dict = {}
        for ds in fds:
            b = ds.bandid
            fds_dict.setdefault(b, [])
            fds_dict[b].append(ds)

        for b, dslist in fds_dict.items():

            log.info(f"Writing movie to {basename}_band{b}_{idfy}.{outfmt}")
            rmss = [ds.rms for ds in dslist]
            medrms = np.median(rmss)
            nframe = len(dslist)
            for i, ds in enumerate(dslist):
                ds.attrs['rms'] = medrms
                ds.attrs['ffrac'] = f'{i}/{nframe}'

            if outfmt == 'gif':
                outim = stream(
                        dslist,
                        renderer=plot_frame,
                        intro_title=f"{basename}-Band{b:04d}",
                        optimize=opts.optimize,
                        processes=True,
                        threads_per_worker=1,
                        fps=opts.fps,
                        max_frames=-1,
                        uri=f'{basename}_band{b}_{idfy}.gif',
                        scratch_dir=f'{scratch_dir}/streamjoy_scratch',
                        # client=client
                    )
            elif outfmt == 'mp4':
                outim = stream(
                        dslist,
                        renderer=plot_frame,
                        intro_title=f"{basename}-Band{b:04d}",
                        write_kwargs={'crf':opts.crf},
                        processes=True,
                        threads_per_worker=1,
                        fps=opts.fps,
                        max_frames=-1,
                        uri=f'{basename}_band{b}_{idfy}.mp4',
                        scratch_dir=f'{scratch_dir}/streamjoy_scratch',
                        # client=client
                    )
            else:
                log.error_and_raise(f"Unsupported format {opts.out_format}", ValueError)

    else:
        log.error_and_raise(f"Can't animate axis {opts.animate_axis}",
                            NotImplementedError)



import numpy as np
from pfb.utils.misc import convolve2gaussres
from pfb.utils.fits import set_wcs, add_beampars, save_fits
from pfb.utils.naming import xds_from_list


def restore_image(ds_name,
                  model_name,
                  residual_name,
                  gaussparf=None,  # final desired resolution
                  nthreads=1):
    '''
    Writes restored image to dds. 
    Final resolution is specified by gaussparf.
    The intrinsic resolution used by default
    '''
    # convolve residual
    if not isinstance(ds_name, list):
        ds_name = [ds_name]

    drop_all_but = [model_name, residual_name, 'PSFPARSN', 'WSUM']
    dds = xds_from_list(ds_name, nthreads=nthreads,
                        drop_all_but=drop_all_but)
    if len(dds) > 1:
        raise RuntimeError('Some thing went wrong. '
                           'This should return a single dataset')
    ds = dds[0]
    if model_name not in ds:
        raise ValueError(f'Could not find {model_name} in dds')
    if residual_name not in ds:
        raise ValueError(f'Could not find {residual_name} in dds')
    wsum = ds.WSUM.values
    model = ds.get(model_name).values
    ncorr, nx, ny = model.shape
    l = -(nx//2) + np.arange(nx)
    m = -(ny//2) + np.arange(ny)
    xx, yy = np.meshgrid(l, m, indexing='ij')
    residual = ds.get(residual_name).values/wsum[:, None, None]

    if gaussparf is not None:
        rconv = convolve2gaussres(residual, xx, yy, gaussparf,
                              nthreads=nthreads,
                              gausspari=ds.PSFPARSN.values,
                              pfrac=0.2,
                              norm_kernel=False)    
    else:
        gaussparf = ds.PSFPARSN.values
        rconv = residual
    mconv = convolve2gaussres(model, xx, yy, gaussparf,
                              nthreads=nthreads,
                              pfrac=0.2,
                              norm_kernel=False)


    

    mconv += rconv

    ds = ds.assign({'IMAGE': (('corr', 'x', 'y'), mconv),
                    'PSFPARSF': (('corr', 'bpar'), gaussparf)})
    # only write updates
    ds = ds[['IMAGE', 'PSFPARSF']]
    ds.to_zarr(ds_name[0], mode='a')



# flake8: noqa
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
import time
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('INIT')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.init)
def init(**kw):
    '''
    Initialise Stokes data products for imaging
    '''
    opts = OmegaConf.create(kw)
    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    remprod = opts.product.upper().strip('IQUV')
    if len(remprod):
        log.error_and_raise(f"Product {remprod} not yet supported",
                            NotImplementedError)

    from daskms.fsspec_store import DaskMSStore
    msnames = []
    for ms in opts.ms:
        msstore = DaskMSStore(ms.rstrip('/'))
        mslist = msstore.fs.glob(ms.rstrip('/'))
        try:
            assert len(mslist) > 0
            msnames.append(*list(map(msstore.fs.unstrip_protocol, mslist)))
        except:
            log.error_and_raise(f"No MS at {ms}", ValueError)
    opts.ms = msnames
    if opts.gain_table is not None:
        gainnames = []
        for gt in opts.gain_table:
            gainstore = DaskMSStore(gt.rstrip('/'))
            gtlist = gainstore.fs.glob(gt.rstrip('/'))
            try:
                assert len(gtlist) > 0
                gainnames.append(*list(map(gainstore.fs.unstrip_protocol, gtlist)))
            except Exception as e:
                log.error_and_raise(f"No gain table  at {gt}", ValueError)
        opts.gain_table = gainnames

    OmegaConf.set_struct(opts, True)

    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/init_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    # with ExitStack() as stack:
    import dask
    dask.config.set(**{'array.slicing.split_large_chunks': False})
    from pfb import set_client
    client = set_client(opts.nworkers, log, client_log_level=opts.log_level)

    ti = time.time()
    _init(**opts)

    log.info(f"All done after {time.time() - ti}s")

    try:
        client.close()
    except Exception as e:
        pass

def _init(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    from pfb.utils.misc import construct_mappings
    from distributed import get_client, as_completed
    from daskms import xds_from_storage_ms as xds_from_ms
    from daskms.experimental.zarr import xds_from_zarr
    from daskms.fsspec_store import DaskMSStore
    from pfb.utils.stokes2vis import single_stokes
    from uuid import uuid4
    import fsspec

    basename = f'{opts.output_filename}'

    xds_store = DaskMSStore(f'{basename}.xds')
    if xds_store.exists():
        if opts.overwrite:
            log.info(f"Overwriting {basename}.xds")
            xds_store.rm(recursive=True)
        else:
            log.error_and_raise(f"{basename}.xds exists. "
                                "Set overwrite to overwrite it. ",
                                RuntimeError)

    fs = fsspec.filesystem(xds_store.protocol)
    fs.makedirs(xds_store.url, exist_ok=True)

    log.info(f"Data products will be stored in {xds_store.url}")

    if opts.gain_table is not None:
        tmpf = lambda x: '::'.join(x.rsplit('/', 1))
        gain_names = list(map(tmpf, opts.gain_table))
    else:
        gain_names = None

    if opts.freq_range is not None and len(opts.freq_range):
        fmin, fmax = opts.freq_range.strip(' ').split(':')
        if len(fmin) > 0:
            freq_min = float(fmin)
        else:
            freq_min = -np.inf
        if len(fmax) > 0:
            freq_max = float(fmax)
        else:
            freq_max = np.inf
    else:
        freq_min = -np.inf
        freq_max = np.inf

    client = get_client()
    worker_keys = client.scheduler_info()['workers'].keys()

    log.info('Constructing mapping')
    row_mapping, freq_mapping, time_mapping, \
        freqs, utimes, ms_chunks, gains, radecs, \
        chan_widths, uv_max, antpos, poltype = \
            construct_mappings(opts.ms,
                               gain_names,
                               ipi=opts.integrations_per_image,
                               cpi=opts.channels_per_image,
                               freq_min=freq_min,
                               freq_max=freq_max,
                               FIELD_IDs=opts.fields,
                               DDIDs=opts.ddids,
                               SCANs=opts.scans)

    group_by = ['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER']

    # crude column arithmetic
    dc = opts.data_column.replace(" ", "")
    if "+" in dc:
        dc1, dc2 = dc.split("+")
        operator="+"
    elif "-" in dc:
        dc1, dc2 = dc.split("-")
        operator="-"
    else:
        dc1 = dc
        dc2 = None
        operator=None

    columns = (dc1,
               'UVW', 'ANTENNA1',
               'ANTENNA2', 'TIME', 'INTERVAL', 'FLAG_ROW')
    schema = {}
    if opts.flag_column != 'None':
        columns += (opts.flag_column,)
        schema[opts.flag_column] = {'dims': ('chan', 'corr')}
    schema[dc1] = {'dims': ('chan', 'corr')}
    if dc2 is not None:
        columns += (dc2,)
        schema[dc2] = {'dims': ('chan', 'corr')}

    # only WEIGHT column gets special treatment
    # any other column must have channel axis
    if opts.sigma_column is not None:
        log.info(f"Initialising weights from {opts.sigma_column} column")
        columns += (opts.sigma_column,)
        schema[opts.sigma_column] = {'dims': ('chan', 'corr')}
    elif opts.weight_column is not None:
        log.info(f"Using weights from {opts.weight_column} column")
        columns += (opts.weight_column,)
        # hack for https://github.com/ratt-ru/dask-ms/issues/268
        if opts.weight_column != 'WEIGHT':
            schema[opts.weight_column] = {'dims': ('chan', 'corr')}
    else:
        log.info(f"No weights provided, using unity weights")

    # distinct freq groups
    sgroup = 0
    freq_groups = []
    freq_sgroups = []
    for ms in opts.ms:
        for idt, freq in freqs[ms].items():
            ilo = idt.find('DDID') + 4
            ihi = idt.rfind('_')
            ddid = int(idt[ilo:ihi])
            if (opts.ddids is not None) and (ddid not in opts.ddids):
                continue
            if not len(freq_groups):
                freq_groups.append(freq)
                freq_sgroups.append(sgroup)
                sgroup += freq_mapping[ms][idt]['counts'].size
            else:
                in_group = False
                for fs in freq_groups:
                    if freq.size == fs.size and np.all(freq == fs):
                        in_group = True
                        break
                if not in_group:
                    freq_groups.append(freq)
                    freq_sgroups.append(sgroup)
                    sgroup += freq_mapping[ms][idt]['counts'].size

    # band mapping
    msddid2bid = {}
    for ms in opts.ms:
        msddid2bid[ms] = {}
        for idt, freq in freqs[ms].items():
            # find group where it matches
            for sgroup, fs in zip(freq_sgroups, freq_groups):
                if freq.size == fs.size and np.all(freq == fs):
                    msddid2bid[ms][idt] = sgroup

    # a flat list to use with as_completed
    datasets = []
    for ims, ms in enumerate(opts.ms):
        xds = xds_from_ms(ms, chunks=ms_chunks[ms], columns=columns,
                          table_schema=schema, group_cols=group_by)

        for ds in xds:
            fid = ds.FIELD_ID
            ddid = ds.DATA_DESC_ID
            scanid = ds.SCAN_NUMBER
            if (opts.fields is not None) and (fid not in opts.fields):
                continue
            if (opts.ddids is not None) and (ddid not in opts.ddids):
                continue
            if (opts.scans is not None) and (scanid not in opts.scans):
                continue

            idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"

            idx = (freqs[ms][idt]>=freq_min) & (freqs[ms][idt]<=freq_max)
            if not idx.any():
                continue

            titr = enumerate(zip(time_mapping[ms][idt]['start_indices'],
                                time_mapping[ms][idt]['counts']))
            for ti, (tlow, tcounts) in titr:

                It = slice(tlow, tlow + tcounts)
                ridx = row_mapping[ms][idt]['start_indices'][It]
                rcnts = row_mapping[ms][idt]['counts'][It]
                # select all rows for output dataset
                Irow = slice(ridx[0], ridx[-1] + rcnts[-1])

                fitr = enumerate(zip(freq_mapping[ms][idt]['start_indices'],
                                     freq_mapping[ms][idt]['counts']))
                b0 = msddid2bid[ms][idt]
                for fi, (flow, fcounts) in fitr:
                    Inu = slice(flow, flow + fcounts)
                    subds = ds[{'row': Irow, 'chan': Inu}]
                    if gains[ms][idt] is not None:
                        subgds = gains[ms][idt][{'gain_time': It, 'gain_freq': Inu}]
                        jones = subgds.gains.data
                    else:
                        jones = None

                    datasets.append([subds,
                                    jones,
                                    freqs[ms][idt][Inu],
                                    chan_widths[ms][idt][Inu],
                                    utimes[ms][idt][It],
                                    ridx, rcnts,
                                    radecs[ms][idt],
                                    b0 + fi, ti, ims, ms, flow, flow+fcounts])

    futures = []
    associated_workers = {}
    idle_workers = set(client.scheduler_info()['workers'].keys())
    n_launched = 0
    nds = len(datasets)
    while idle_workers and len(datasets):   # Seed each worker with a task.
        # pop so len(datasets) -> 0
        (subds, jones, freqsi, chan_widthi, utimesi, ridx, rcnts,
         radeci, fi, ti, ims, ms, chan_low, chan_high) = datasets.pop(0)

        worker = idle_workers.pop()
        future = client.submit(single_stokes,
                        dc1=dc1,
                        dc2=dc2,
                        operator=operator,
                        ds=subds,
                        jones=jones,
                        opts=opts,
                        freq=freqsi,
                        chan_width=chan_widthi,
                        utime=utimesi,
                        tbin_idx=ridx,
                        tbin_counts=rcnts,
                        chan_low=chan_low,
                        chan_high=chan_high,
                        radec=radeci,
                        antpos=antpos[ms],
                        poltype=poltype[ms],
                        xds_store=xds_store.url,
                        bandid=fi,
                        timeid=ti,
                        msid=ims,
                        wid=worker,
                        pure=False,
                        workers=worker,
                        key='image-'+uuid4().hex)

        futures.append(future)
        associated_workers[future] = worker
        n_launched += 1

        if opts.progressbar:
            print(f"\rProcessing: {n_launched}/{nds}", end='', flush=True)

    times_out = []
    freqs_out = []
    ac_iter = as_completed(futures)
    for completed_future in ac_iter:
        if isinstance(completed_future.result(), BaseException):
            e = completed_future.result()
            import traceback
            log.error_and_raise(f"Traceback:\n{traceback.format_exc()}", e)

        result = completed_future.result()
        if result is not None:
            times_out.append(result[0])
            freqs_out.append(result[1])

        worker = associated_workers.pop(completed_future)
        # we need this here to release memory for some reason
        client.cancel(completed_future)

        # pop so len(datasets) -> 0
        if len(datasets):
            (subds, jones, freqsi, chan_widthi, utimesi, ridx, rcnts,
            radeci, fi, ti, ims, ms, chan_low, chan_high) = datasets.pop(0)

            future = client.submit(single_stokes,
                            dc1=dc1,
                            dc2=dc2,
                            operator=operator,
                            ds=subds,
                            jones=jones,
                            opts=opts,
                            freq=freqsi,
                            chan_width=chan_widthi,
                            utime=utimesi,
                            tbin_idx=ridx,
                            tbin_counts=rcnts,
                            chan_low=chan_low,
                            chan_high=chan_high,
                            radec=radeci,
                            antpos=antpos[ms],
                            poltype=poltype[ms],
                            xds_store=xds_store.url,
                            bandid=fi,
                            timeid=ti,
                            msid=ims,
                            wid=worker,
                            pure=False,
                            workers=worker,
                            key='image-'+uuid4().hex)


            ac_iter.add(future)
            associated_workers[future] = worker
            n_launched += 1

        if opts.memory_reporting:
            worker_info = client.scheduler_info()['workers']
            log.info(f'Total memory {worker} MB = ',
                  worker_info[worker]['metrics']['memory']/1e6)

        if opts.progressbar:
            print(f"\rProcessing: {n_launched}/{nds}", end='', flush=True)

        # this should not be necessary but just in case
        if ac_iter.is_empty():
            break

    times_out = np.unique(times_out)
    freqs_out = np.unique(freqs_out)

    nband = freqs_out.size
    ntime = times_out.size

    print("\n")  # after progressbar above
    log.info(f"Freq and time selection resulted in {nband} output bands and "
          f"{ntime} output times")

    return



# flake8: noqa
from contextlib import ExitStack
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('SPOTLESS')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.spotless)
def spotless(**kw):
    '''
    Distributed spotless algorithm.
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    from daskms.fsspec_store import DaskMSStore
    from pfb.utils.naming import xds_from_url

    basename = f'{basedir}/{oname}'
    if opts.xds is not None:
        xds_store = DaskMSStore(opts.xds.rstrip('/'))
        xds_name = opts.xds
    else:
        xds_name = f'{basename}.xds'
        xds_store = DaskMSStore(xds_name)
        opts.xds = xds_name
    fits_oname = f'{opts.fits_output_folder}/{oname}'
    dds_name = f'{basename}_{opts.suffix}.dds'

    OmegaConf.set_struct(opts, True)

    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{opts.log_directory}/spotless_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    with ExitStack() as stack:
        from pfb import set_client
        if opts.nworkers > 1:
            client = set_client(opts.nworkers, log, stack=stack,
                                direct_to_workers=opts.direct_to_workers,
                                client_log_level=opts.log_level)
            from distributed import as_completed
        else:
            log.info("Faking client")
            from pfb.utils.dist import fake_client
            client = fake_client()
            names = [0]
            as_completed = lambda x: x

        ti = time.time()
        _spotless(**opts)

        dds, dds_list = xds_from_url(dds_name)

        # convert to fits files
        futures = []
        if opts.fits_mfs or opts.fits_cubes:
            from pfb.utils.fits import dds2fits
            log.info(f"Writing fits files to {fits_oname}_{opts.suffix}")
            fut = client.submit(dds2fits,
                                dds_list,
                                'MODEL',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=False,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes)
            futures.append(fut)

            fut = client.submit(dds2fits,
                                dds_list,
                                'UPDATE',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=False,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes)
            futures.append(fut)

            fut = client.submit(dds2fits,
                                dds_list,
                                'RESIDUAL',
                                f'{fits_oname}_{opts.suffix}',
                                norm_wsum=True,
                                nthreads=opts.nthreads,
                                do_mfs=opts.fits_mfs,
                                do_cube=opts.fits_cubes)
            futures.append(fut)

            for fut in as_completed(futures):
                column = fut.result()
                log.info(f'Done writing {column}')

    log.info(f"All done after {time.time() - ti}s.")


def _spotless(**kw):
    '''
    Distributed spotless algorithm.

    The key inputs to the algorithm are the xds and mds.
    The xds contains the Stokes coherencies created with the init worker.
    There can be multiple datasets in the xds, each corresponding to a specific frequency and time range.
    These datasets are persisted on separate workers.
    The mds contains a continuous representation of the model in compressed format.
    The mds is just a single dataset which gets evaluated into a discretised cube on the runner node.
    The frequency resolution of the model is the same as the frequency resoltuion of the xds.

    '''
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    from itertools import cycle
    import numpy as np
    import numba
    import xarray as xr
    from distributed import get_client
    from pfb.opt.power_method import power_method_dist as power_method
    from pfb.opt.primal_dual import primal_dual_dist as primal_dual
    from pfb.utils.naming import xds_from_url, get_opts, cache_opts
    from pfb.utils.fits import load_fits, dds2fits, set_wcs, save_fits
    from pfb.utils.dist import l1reweight_func
    from itertools import cycle
    from uuid import uuid4
    from daskms.fsspec_store import DaskMSStore
    from pfb.utils.modelspec import eval_coeffs_to_slice, fit_image_cube
    import fsspec
    import pickle
    from pfb.utils.dist import band_actor

    # the runner should use all available resources
    import psutil
    ncpu = psutil.cpu_count(logical=False)
    numba.set_num_threads(ncpu)

    real_type = np.float64
    complex_type = np.complex128

    try:
        client = get_client()
        names = list(client.scheduler_info()['workers'].keys())
    except:
        from pfb.utils.dist import fake_client
        client = fake_client()
        names = [0]
        as_completed = lambda x: x

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    # xds contains vis products, no imaging weights applied
    xds_name = f'{basename}.xds' if opts.xds is None else opts.xds
    xds_store = DaskMSStore(xds_name)
    try:
        assert xds_store.exists()
    except Exception as e:
        log.error_and_raise(f"There must be a dataset at {xds_store.url}",
                            ValueError)

    xds, xds_list = xds_from_url(xds_store.url)

    # create dds and cache
    dds_name = opts.output_filename + f'_{opts.suffix}' + '.dds'
    dds_store = DaskMSStore(dds_name)
    if '://' in dds_store.url:
        protocol = xds_store.url.split('://')[0]
    else:
        protocol = 'file'

    if dds_store.exists() and opts.overwrite:
        log.info(f"Removing {dds_store.url}")
        dds_store.rm(recursive=True)

    optsp_name = dds_store.url + '/opts.pkl'
    fs = fsspec.filesystem(protocol)
    if dds_store.exists() and not opts.overwrite:
        # get opts from previous run
        optsp = get_opts(dds_store.url,
                         protocol,
                         name='opts.pkl')

        # check if we need to remake the data products
        verify_attrs = ['epsilon',
                        'do_wgridding', 'double_accum',
                        'field_of_view', 'super_resolution_factor']
        try:
            for attr in verify_attrs:
                assert optsp[attr] == opts[attr]

            from_cache = True
            log.info("Initialising from cached data products")
            dds, dds_list = xds_from_url(dds_store.url)
            iter0 = dds[0].niters
        except Exception as e:
            log.info(f'Cache verification failed on {attr}. '
                  'Will remake image data products')
            dds_store.rm(recursive=True)
            fs.makedirs(dds_store.url, exist_ok=True)
            # dump opts to validate cache on rerun
            cache_opts(opts,
                       dds_store.url,
                       protocol,
                       name='opts.pkl')
            from_cache = False
            iter0 = 0
    else:
        fs.makedirs(dds_store.url, exist_ok=True)
        cache_opts(opts,
                   dds_store.url,
                   protocol,
                   name='opts.pkl')
        from_cache = False
        log.info("Initialising from scratch.")
        iter0 = 0

    log.info(f"Image data products will be cached in {dds_store.url}")

    uv_max = 0
    max_freq = 0
    for ds in xds:
        uv_max = np.maximum(uv_max, ds.uv_max)
        max_freq = np.maximum(max_freq, ds.max_freq)

    # filter datasets by band
    xdsb = {}
    for ds in xds_list:
        idx = ds.find('band') + 4
        bid = int(ds[idx:idx+4])
        xdsb.setdefault(bid, [])
        xdsb[bid].append(ds)

    nband = len(xdsb.keys())

    # set up band actors
    log.info("Setting up actors")
    futures = []
    for wname, (bandid, dsl) in zip(cycle(names), xdsb.items()):
        f = client.submit(band_actor,
                          dsl,
                          opts,
                          bandid,
                          dds_store.url,
                          uv_max,
                          max_freq,
                          workers=wname,
                          key='actor-'+uuid4().hex,
                          actor=True,
                          pure=False)
        futures.append(f)
    actors = list(map(lambda f: f.result(), futures))
    futures = list(map(lambda a: a.get_image_info(), actors))
    results = list(map(lambda f: f.result(), futures))
    nx, ny, nymax, nxmax, cell_rad, ra, dec, x0, y0, freq_out, time_out = results[0]
    freq_out = [freq_out]
    for result in results[1:]:
        assert result[0] == nx
        assert result[1] == ny
        assert result[2] == nymax
        assert result[3] == nxmax
        assert result[4] == cell_rad
        assert result[5] == ra
        assert result[6] == dec
        assert result[7] == x0
        assert result[8] == y0
        freq_out.append(result[9])

    log.info(f"Image size set to ({nx}, {ny})")

    radec = [ra, dec]
    cell_deg = np.rad2deg(cell_rad)
    freq_out = np.array(freq_out)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)
    time_out = np.array((time_out,))

    # get size of dual domain
    bases = tuple(opts.bases.split(','))
    nbasis = len(bases)

    if opts.mds is not None:
        mds_name = opts.mds
    else:
        mds_name = opts.output_filename + f'_{opts.suffix}' + '.mds'

    mds_store = DaskMSStore(mds_name)

    # initialise model (assumed constant in time)
    if mds_store.exists():
        log.info(f"Loading model from {mds_store.url}")
        mds = xr.open_zarr(mds_store.url, chunks=None)

        # we only want to load these once
        model_coeffs = mds.coefficients.values
        locx = mds.location_x.values
        locy = mds.location_y.values
        params = mds.params.values
        coeffs = mds.coefficients.values

        model = np.zeros((nband, nx, ny))
        for i in range(nband):
            model[i] = eval_coeffs_to_slice(
                time_out[0],
                freq_out[i],
                model_coeffs,
                locx, locy,
                mds.parametrisation,
                params,
                mds.texpr,
                mds.fexpr,
                mds.npix_x, mds.npix_y,
                mds.cell_rad_x, mds.cell_rad_y,
                mds.center_x, mds.center_y,
                nx, ny,
                cell_rad, cell_rad,
                x0, y0
            )

    else:
        model = np.zeros((nband, nx, ny))

    log.info("Computing image data products")
    futures = []
    for b in range(nband):
        fut = actors[b].set_image_data_products(model[b],
                                                iter0,
                                                from_cache=from_cache)
        futures.append(fut)


    results = list(map(lambda f: f.result(), futures))
    residual_mfs = np.sum([r[0] for r in results], axis=0)
    wsums = np.array([r[1] for r in results])
    wsum = np.sum(wsums)
    fsel = wsums > 0
    futures = list(map(lambda a: a.set_wsum(wsum), actors))
    results = list(map(lambda f: f.result(), futures))

    residual_mfs /= wsum
    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()

    if iter0 == 0:
        save_fits(residual_mfs,
                  fits_oname + f'_{opts.suffix}_dirty_mfs.fits',
                  hdr_mfs)

    if opts.hess_norm is None:
        log.info('Getting spectral norm of Hessian approximation')
        hess_norm = power_method(actors, nx, ny, nband,
                                tol=opts.pm_tol,
                                maxit=opts.pm_maxit,
                                report_freq=opts.pm_report_freq,
                                verbosity=opts.pm_verbose)
        # inflate so we don't have to recompute after each L2 reweight
        hess_norm *= 1.05
    else:
        hess_norm = opts.hess_norm
    log.info(f'hess-norm = {hess_norm:.3e}')

    # a value less than zero turns L1 reweighting off
    # we'll start on convergence or at the iteration
    # indicated by l1_reweight_from, whichever comes first
    l1_reweight_from = opts.l1_reweight_from
    l1reweight_active = False

    if l1_reweight_from == 0:
        log.info(f'Initialising with L1 reweighted')
        l1weight, rms_comps = l1reweight_func(actors,
                                   opts.rmsfactor,
                                   alpha=opts.alpha)
        l1reweight_active = True
    else:
        rms_comps = rms  # this should not happen
        l1weight = np.ones((nbasis, nymax, nxmax), dtype=real_type)
        l1reweight_active = False

    best_rms = rms
    best_rmax = rmax
    best_model = model.copy()
    diverge_count = 0
    l2_reweights = 0
    log.info(f"It {iter0}: max resid = {rmax:.3e}, rms = {rms:.3e}")
    for k in range(iter0, iter0 + opts.niter):
        log.info('Solving for update')
        futures = list(map(lambda a: a.cg_update(), actors))
        update_mfs = np.sum(list(map(lambda f: f.result(), futures)), axis=0)
        update_mfs /= np.sum(fsel)

        save_fits(update_mfs,
                  fits_oname + f'_{opts.suffix}_update_{k+1}.fits',
                  hdr_mfs)

        log.info('Solving for model')
        primal_dual(actors,
                    rms*opts.rmsfactor,
                    hess_norm,
                    l1weight,
                    opts.rmsfactor,
                    rms_comps,
                    opts.alpha,
                    nu=len(bases),
                    tol=opts.pd_tol,
                    maxit=opts.pd_maxit,
                    positivity=opts.positivity,
                    gamma=opts.gamma,
                    verbosity=opts.pd_verbose)

        log.info('Updating model')
        futures = list(map(lambda a: a.give_model(), actors))
        results = list(map(lambda f: f.result(), futures))

        bandids = [r[1] for r in results]
        modelp = model.copy()
        for i, b in enumerate(bandids):
            model[b] = results[i][0]

        if np.isnan(model).any():
            log.error_and_raise('Model is nan', RuntimeError)

        save_fits(np.mean(model[fsel], axis=0),
                  fits_oname + f'_{opts.suffix}_model_{k+1}.fits',
                  hdr_mfs)

        log.info(f"Writing model to {mds_store.url}")
        coeffs, Ix, Iy, expr, params, texpr, fexpr = \
            fit_image_cube(time_out, freq_out[fsel], model[None, fsel, :, :],
                            wgt=wsums[None, fsel],
                            method='Legendre')

        data_vars = {
            'coefficients': (('par', 'comps'), coeffs),
        }
        coords = {
            'location_x': (('x',), Ix),
            'location_y': (('y',), Iy),
            'params': (('par',), params),  # already converted to list
            'times': (('t',), time_out),  # to allow rendering to original grid
            'freqs': (('f',), freq_out)
        }
        attrs = {
            'spec': 'genesis',
            'cell_rad_x': cell_rad,
            'cell_rad_y': cell_rad,
            'npix_x': nx,
            'npix_y': ny,
            'texpr': texpr,
            'fexpr': fexpr,
            'center_x': x0,
            'center_y': y0,
            'ra': ra,
            'dec': dec,
            'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
            'parametrisation': expr,  # already converted to str
        }

        coeff_dataset = xr.Dataset(data_vars=data_vars,
                            coords=coords,
                            attrs=attrs)
        coeff_dataset.to_zarr(f"{mds_store.url}", mode='w')

        # convergence check
        eps = np.linalg.norm(model - modelp)/np.linalg.norm(model)
        if eps < opts.tol:
            # do not converge prematurely
            if l1_reweight_from > 0 and not l1reweight_active:  # only happens once
                # start reweighting
                l1_reweight_from = k+1 - iter0
                # don't start L2 reweighting before L1 reweighting
                # if it is enabled
                dof = None
                l1reweight_active = True
            elif l2_reweights < opts.max_l2_reweight:
                # L1 reweighting has already kicked in and we have
                # converged again so perform L2 reweight
                dof = opts.l2_reweight_dof
            else:
                # we have reached max L2 reweights so we are done
                log.info(f"Converged after {k+1} iterations.")
                break
        else:
            # do not perform an L2 reweight unless the M step has converged
            dof = None

        if k+1 - iter0 >= opts.l2_reweight_from and l2_reweights < opts.max_l2_reweight:
            dof = opts.l2_reweight_dof

        if dof is not None:
            log.info('Recomputing image data products since L2 reweight '
                  'is required.')
            futures = []
            for b in range(nband):
                fut = actors[b].set_image_data_products(model[b],
                                                        k+1,
                                                        dof=dof)
                futures.append(fut)


            results = list(map(lambda f: f.result(), futures))
            residual_mfs = np.sum([r[0] for r in results], axis=0)
            wsums = np.array([r[1] for r in results])
            wsum = np.sum(wsums)
            futures = list(map(lambda a: a.set_wsum(wsum), actors))
            results = list(map(lambda f: f.result(), futures))

            residual_mfs /= wsum

            # don't keep the cubes on the runner
            del results

            # # TODO - how much does hess-norm change after L2 reweight?
            # log.info('Getting spectral norm of Hessian approximation')
            # hess_norm = power_method(actors, nx, ny, nband)
            # log.info(f'hess-norm = {hess_norm:.3e}')

            l2_reweights += 1
        else:
            # compute normal residual, no need to redo PSF etc
            log.info('Computing residual')
            futures = list(map(lambda a: a.set_residual(k+1), actors))
            resids = list(map(lambda f: f.result(), futures))
            # we never resids by wsum inside the worker
            residual_mfs = np.sum(resids, axis=0)/wsum


        save_fits(residual_mfs,
                  fits_oname + f'_{opts.suffix}_residual_{k+1}.fits',
                  hdr_mfs)

        rmsp = rms
        rms = np.std(residual_mfs)
        rmax = np.abs(residual_mfs).max()

        # base this on rmax?
        if rms < best_rms:
            best_rms = rms
            best_rmax = rmax
            best_model = model.copy()

        log.info(f"It {k+1}: max resid = {rmax:.3e}, rms = {rms:.3e}, eps = {eps:.3e}")

        if k+1 - iter0 >= l1_reweight_from:
            log.info('L1 reweighting')
            l1weight, rms_comps = l1reweight_func(actors,
                                       opts.rmsfactor,
                                       alpha=opts.alpha)

        if rms > rmsp:
            diverge_count += 1
            if diverge_count > opts.diverge_count:
                log.info("Algorithm is diverging. Terminating.")
                break

    return

